from __future__ import annotations

__all__ = ["PoissonTensorGenerator", "RandPoissonTensorGenerator"]


import torch
from coola.utils.format import str_indent, str_mapping
from torch import Generator, Tensor

from startorch.random import rand_poisson
from startorch.tensor.base import BaseTensorGenerator, setup_tensor_generator


class PoissonTensorGenerator(BaseTensorGenerator):
    r"""Implements a class to generate tensors by sampling values from a
    Poisson distribution.

    The rates of the Poisson distribution are generated by the rate
    generator. The rate generator should return the rate for each value
    in the tensor. The rate values should be greater than 0.

    Args:
    ----
        rate (``BaseTensorGenerator`` or dict): Specifies the
            rate generator or its configuration. The rate generator
            should return valid rate values.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.tensor import RandUniform, Poisson
        >>> generator = Poisson(rate=RandUniform(low=1.0, high=2.0))
        >>> generator
        PoissonTensorGenerator(
          (rate): RandUniformTensorGenerator(low=1.0, high=2.0)
        )
        >>> generator.generate((2, 6))
        tensor([[...]])
    """

    def __init__(self, rate: BaseTensorGenerator | dict) -> None:
        super().__init__()
        self._rate = setup_tensor_generator(rate)

    def __repr__(self) -> str:
        args = str_indent(str_mapping({"rate": self._rate}))
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def generate(self, size: tuple[int, ...], rng: Generator | None = None) -> Tensor:
        return torch.poisson(self._rate.generate(size=size, rng=rng), generator=rng)


class RandPoissonTensorGenerator(BaseTensorGenerator):
    r"""Implements a class to generate tensors by sampling values from a
    Poisson distribution.

    Args:
    ----
        rate (float, optional): Specifies the rate of the Poisson
            distribution. This value has to be greater than 0.
            Default: ``1.0``

    Raises:
    ------
        ValueError if ``rate`` is not a positive number.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.tensor import RandPoisson
        >>> generator = RandPoisson(rate=1.0)
        >>> generator
        RandPoissonTensorGenerator(rate=1.0)
        >>> generator.generate((2, 6))
        tensor([[...]])
    """

    def __init__(self, rate: float = 1.0) -> None:
        super().__init__()
        if rate <= 0:
            raise ValueError(f"rate has to be greater than 0 (received: {rate})")
        self._rate = float(rate)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(rate={self._rate})"

    def generate(self, size: tuple[int, ...], rng: Generator | None = None) -> Tensor:
        return rand_poisson(size=size, rate=self._rate, generator=rng)



from __future__ import annotations

__all__ = ["BaseWrapperSequenceGenerator"]


from coola.utils import str_indent, str_mapping

from startorch.sequence.base import BaseSequenceGenerator, setup_sequence_generator


class BaseWrapperSequenceGenerator(BaseSequenceGenerator):
    r"""Defines a base class to easily wrap a sequence generator.

    Args:
    ----
        generator (``BaseSequenceGenerator`` or dict):
            Specifies the sequence generator or its configuration.
    """

    def __init__(self, generator: BaseSequenceGenerator | dict) -> None:
        super().__init__()
        self._generator = setup_sequence_generator(generator)

    def __repr__(self) -> str:
        args = str_indent(str_mapping({"sequence": self._generator}))
        return f"{self.__class__.__qualname__}(\n  {args}\n)"



from __future__ import annotations

__all__ = [
    "HalfNormalSequenceGenerator",
    "RandHalfNormalSequenceGenerator",
    "RandTruncHalfNormalSequenceGenerator",
    "TruncHalfNormalSequenceGenerator",
]


from coola.utils.format import str_indent, str_mapping
from redcat import BatchedTensorSeq
from torch import Generator

from startorch.random import (
    half_normal,
    rand_half_normal,
    rand_trunc_half_normal,
    trunc_half_normal,
)
from startorch.sequence.base import BaseSequenceGenerator, setup_sequence_generator
from startorch.utils.conversion import to_tuple


class HalfNormalSequenceGenerator(BaseSequenceGenerator):
    r"""Implements a class to generate sequence by sampling values from a
    half-Normal distribution.

    Args:
    ----
        std (``BaseSequenceGenerator`` or dict): Specifies a sequence
            generator (or its configuration) to generate the standard
            deviation.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.sequence import HalfNormal, RandUniform
        >>> generator = HalfNormal(std=RandUniform(low=1.0, high=2.0))
        >>> generator
        HalfNormalSequenceGenerator(
          (std): RandUniformSequenceGenerator(low=1.0, high=2.0, feature_size=(1,))
        )
        >>> generator.generate(seq_len=6, batch_size=2)
        tensor([[...]], batch_dim=0, seq_dim=1)
    """

    def __init__(self, std: BaseSequenceGenerator | dict) -> None:
        super().__init__()
        self._std = setup_sequence_generator(std)

    def __repr__(self) -> str:
        args = str_indent(str_mapping({"std": self._std}))
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def generate(
        self, seq_len: int, batch_size: int = 1, rng: Generator | None = None
    ) -> BatchedTensorSeq:
        return BatchedTensorSeq(
            half_normal(
                std=self._std.generate(seq_len=seq_len, batch_size=batch_size, rng=rng).data,
                generator=rng,
            )
        )


class RandHalfNormalSequenceGenerator(BaseSequenceGenerator):
    r"""Implements a class to generate sequence by sampling values from a
    half-Normal distribution.

    Args:
    ----
        std (float, optional): Specifies the std of the distribution.
            Default: ``1.0``
        feature_size (tuple or list or int, optional): Specifies the
            feature size. Default: ``1``

    Raises:
    ------
        ValueError if ``std`` is not a positive number.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.sequence import RandHalfNormal
        >>> generator = RandHalfNormal(std=1.0)
        >>> generator
        RandHalfNormalSequenceGenerator(std=1.0, feature_size=(1,))
        >>> generator.generate(seq_len=6, batch_size=2)
        tensor([[...]], batch_dim=0, seq_dim=1)
    """

    def __init__(
        self, std: float = 1.0, feature_size: tuple[int, ...] | list[int] | int = 1
    ) -> None:
        super().__init__()
        if std <= 0:
            raise ValueError(f"std has to be greater than 0 (received: {std})")
        self._std = float(std)
        self._feature_size = to_tuple(feature_size)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(std={self._std}, feature_size={self._feature_size})"

    def generate(
        self, seq_len: int, batch_size: int = 1, rng: Generator | None = None
    ) -> BatchedTensorSeq:
        return BatchedTensorSeq(
            rand_half_normal(
                size=(batch_size, seq_len) + self._feature_size,
                std=self._std,
                generator=rng,
            )
        )


class RandTruncHalfNormalSequenceGenerator(BaseSequenceGenerator):
    r"""Implements a class to generate sequence by sampling values from a
    truncated half-Normal distribution.

    Args:
    ----
        std (float, optional): Specifies the std of the distribution.
            Default: ``1.0``
        max_value (float, optional): Specifies the maximum value.
            Default: ``3.0``
        feature_size (tuple or list or int, optional): Specifies the
            feature size. Default: ``1``

    Raises:
    ------
        ValueError if ``std`` is not a positive number.
        ValueError if ``max_value`` is not a positive number.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.sequence import RandTruncHalfNormal
        >>> generator = RandTruncHalfNormal(std=1.0, max_value=1.0)
        >>> generator
        RandTruncHalfNormalSequenceGenerator(std=1.0, max_value=1.0, feature_size=(1,))
        >>> generator.generate(seq_len=6, batch_size=2)
        tensor([[...]], batch_dim=0, seq_dim=1)
    """

    def __init__(
        self,
        std: float = 1.0,
        max_value: float = 3.0,
        feature_size: tuple[int, ...] | list[int] | int = 1,
    ) -> None:
        super().__init__()
        if std <= 0:
            raise ValueError(f"std has to be greater than 0 (received: {std})")
        self._std = float(std)
        if max_value <= 0:
            raise ValueError(f"max_value has to be greater than 0 (received: {max_value})")
        self._max_value = float(max_value)
        self._feature_size = to_tuple(feature_size)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(std={self._std}, max_value={self._max_value}, "
            f"feature_size={self._feature_size})"
        )

    def generate(
        self, seq_len: int, batch_size: int = 1, rng: Generator | None = None
    ) -> BatchedTensorSeq:
        return BatchedTensorSeq(
            rand_trunc_half_normal(
                size=(batch_size, seq_len) + self._feature_size,
                std=self._std,
                max_value=self._max_value,
                generator=rng,
            )
        )


class TruncHalfNormalSequenceGenerator(BaseSequenceGenerator):
    r"""Implements a class to generate sequence by sampling values from a
    half-Normal distribution.

    Args:
    ----
        std (``BaseSequenceGenerator`` or dict): Specifies a sequence
            generator (or its configuration) to generate the std.
        max_value (``BaseSequenceGenerator`` or dict): Specifies a
            sequence generator (or its configuration) to generate the
            maximum value (excluded).

    Example usage:

    .. code-block:: pycon

        >>> from startorch.sequence import RandUniform, TruncHalfNormal
        >>> generator = TruncHalfNormal(
        ...     std=RandUniform(low=1.0, high=2.0),
        ...     max_value=RandUniform(low=5.0, high=10.0),
        ... )
        >>> generator
        TruncHalfNormalSequenceGenerator(
          (std): RandUniformSequenceGenerator(low=1.0, high=2.0, feature_size=(1,))
          (max_value): RandUniformSequenceGenerator(low=5.0, high=10.0, feature_size=(1,))
        )
        >>> generator.generate(seq_len=6, batch_size=2)
        tensor([[...]], batch_dim=0, seq_dim=1)
    """

    def __init__(
        self, std: BaseSequenceGenerator | dict, max_value: BaseSequenceGenerator | dict
    ) -> None:
        super().__init__()
        self._std = setup_sequence_generator(std)
        self._max_value = setup_sequence_generator(max_value)

    def __repr__(self) -> str:
        args = str_indent(str_mapping({"std": self._std, "max_value": self._max_value}))
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def generate(
        self, seq_len: int, batch_size: int = 1, rng: Generator | None = None
    ) -> BatchedTensorSeq:
        return BatchedTensorSeq(
            trunc_half_normal(
                std=self._std.generate(seq_len=seq_len, batch_size=batch_size, rng=rng).data,
                max_value=self._max_value.generate(
                    seq_len=seq_len, batch_size=batch_size, rng=rng
                ).data,
                generator=rng,
            )
        )



from __future__ import annotations

__all__ = ["HypercubeExampleGenerator", "create_hypercube_examples"]

import torch
from redcat import BatchDict, BatchedTensor

from startorch import constants as ct
from startorch.example.base import BaseExampleGenerator


class HypercubeExampleGenerator(BaseExampleGenerator[BatchedTensor]):
    r"""Implements a classification example generator.

    The data are generated by using a hypercube. The targets are some
    vertices of the hypercube. Each input feature is a 1-hot
    representation of the target plus a Gaussian noise. These data can
    be used for a multi-class classification task.

    Args:
    ----
        num_classes (int, optional): Specifies the number of classes.
            Default: 50
        feature_size (int, optional): Specifies the feature size.
            The feature size has to be greater than the number of
            classes. Default: ``64``
        noise_std (float, optional): Specifies the standard deviation
            of the Gaussian noise. Default: ``0.2``

    Raises:
    ------
        ValueError if one of the parameters is not valid.


    Example usage:

    .. code-block:: pycon

        >>> from startorch.example import Hypercube
        >>> generator = Hypercube(num_classes=5, feature_size=6)
        >>> generator
        HypercubeExampleGenerator(num_classes=5, feature_size=6, noise_std=0.2)
        >>> batch = generator.generate(batch_size=10)
        >>> batch
        BatchDict(
          (target): tensor([...], batch_dim=0)
          (feature): tensor([[...]], batch_dim=0)
        )
    """

    def __init__(
        self,
        num_classes: int = 50,
        feature_size: int = 64,
        noise_std: float = 0.2,
    ) -> None:
        if num_classes < 1:
            raise ValueError(f"The number of classes ({num_classes}) has to be greater than 0")
        self._num_classes = int(num_classes)

        if feature_size < num_classes:
            raise ValueError(
                f"The feature dimension ({feature_size:,}) has to be greater or equal to the "
                f"number of classes ({num_classes:,})"
            )
        self._feature_size = int(feature_size)

        if noise_std < 0:
            raise ValueError(
                f"The standard deviation of the Gaussian noise ({noise_std}) has to be "
                "greater or equal than 0"
            )
        self._noise_std = float(noise_std)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}("
            f"num_classes={self._num_classes:,}, "
            f"feature_size={self._feature_size:,}, "
            f"noise_std={self._noise_std:,})"
        )

    @property
    def num_classes(self) -> int:
        r"""``int``: The number of classes when the data are created."""
        return self._num_classes

    @property
    def feature_size(self) -> int:
        r"""``int``: The feature size when the data are created."""
        return self._feature_size

    @property
    def noise_std(self) -> float:
        r"""``float``: The standard deviation of the Gaussian noise."""
        return self._noise_std

    def generate(
        self, batch_size: int = 1, rng: torch.Generator | None = None
    ) -> BatchDict[BatchedTensor]:
        return create_hypercube_examples(
            num_examples=batch_size,
            num_classes=self._num_classes,
            feature_size=self._feature_size,
            noise_std=self._noise_std,
            generator=rng,
        )


def create_hypercube_examples(
    num_examples: int = 1000,
    num_classes: int = 50,
    feature_size: int = 64,
    noise_std: float = 0.2,
    generator: torch.Generator | None = None,
) -> BatchDict[BatchedTensor]:
    r"""Create a toy classification dataset based on hypercube vertex
    structure.

    The data are generated by using a hypercube. The targets are some
    vertices of the hypercube. Each input feature is a 1-hot
    representation of the target plus a Gaussian noise. These data can
    be used for a multi-class classification task.

    Args:
    ----
        num_examples (int, optional): Specifies the number of examples.
            Default: ``1000``
        num_classes (int, optional): Specifies the number of classes.
            Default: 50
        feature_size (int, optional): Specifies the feature size.
            The feature size has to be greater than the number of
            classes. Default: ``64``
        noise_std (float, optional): Specifies the standard deviation
            of the Gaussian noise. Default: ``0.2``
        generator (``torch.Generator`` or ``None``, optional):
            Specifies an optional random generator. Default: ``None``

    Returns:
    -------
        dict: A dictionary with two keys:
            - ``'input'``: a ``torch.Tensor`` of type float and
                shape ``(num_examples, feature_size)``. This
                tensor represents the input features.
            - ``'target'``: a ``torch.Tensor`` of type long and
                shape ``(num_examples,)``. This tensor represents
                the targets.

    Raises:
    ------
        RuntimeError if one of the parameters is not valid.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.example.hypercube import create_hypercube_examples
        >>> batch = create_hypercube_examples(num_examples=10, num_classes=5, feature_size=10)
        >>> batch
        BatchDict(
          (target): tensor([...], batch_dim=0)
          (feature): tensor([[...]], batch_dim=0)
        )
    """

    if num_examples < 1:
        raise RuntimeError(f"The number of examples ({num_examples}) has to be greater than 0")
    if num_classes < 1:
        raise RuntimeError(f"The number of classes ({num_classes}) has to be greater than 0")
    if feature_size < num_classes:
        raise RuntimeError(
            f"The feature dimension ({feature_size:,}) has to be greater or equal to the "
            f"number of classes ({num_classes:,})"
        )
    if noise_std < 0:
        raise RuntimeError(
            f"The standard deviation of the Gaussian noise ({noise_std}) has to be "
            "greater or equal than 0"
        )
    # Generate the target of each example.
    targets = torch.randint(0, num_classes, (num_examples,), generator=generator)
    # Generate the features. Each class should be a vertex of the hyper-cube
    # plus Gaussian noise.
    features = torch.randn(num_examples, feature_size, generator=generator).mul(noise_std)
    features.scatter_add_(1, targets.view(num_examples, 1), torch.ones(num_examples, 1))
    return BatchDict({ct.TARGET: BatchedTensor(targets), ct.FEATURE: BatchedTensor(features)})



from __future__ import annotations

__all__ = [
    "NormalSequenceGenerator",
    "RandNormalSequenceGenerator",
    "RandTruncNormalSequenceGenerator",
    "TruncNormalSequenceGenerator",
]


from coola.utils.format import str_indent, str_mapping
from redcat import BatchedTensorSeq
from torch import Generator

from startorch.random import normal, rand_normal, rand_trunc_normal, trunc_normal
from startorch.sequence.base import BaseSequenceGenerator, setup_sequence_generator
from startorch.utils.conversion import to_tuple


class NormalSequenceGenerator(BaseSequenceGenerator):
    r"""Implements a class to generate sequence by sampling values from a
    Normal distribution.

    Args:
    ----
        mean (``BaseSequenceGenerator`` or dict): Specifies a sequence
            generator (or its configuration) to generate the mean.
        std (``BaseSequenceGenerator`` or dict): Specifies a sequence
            generator (or its configuration) to generate the standard
            deviation.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.sequence import Normal, RandUniform
        >>> generator = Normal(
        ...     mean=RandUniform(low=-1.0, high=1.0), std=RandUniform(low=1.0, high=2.0)
        ... )
        >>> generator
        NormalSequenceGenerator(
          (mean): RandUniformSequenceGenerator(low=-1.0, high=1.0, feature_size=(1,))
          (std): RandUniformSequenceGenerator(low=1.0, high=2.0, feature_size=(1,))
        )
        >>> generator.generate(seq_len=6, batch_size=2)
        tensor([[...]], batch_dim=0, seq_dim=1)
    """

    def __init__(
        self, mean: BaseSequenceGenerator | dict, std: BaseSequenceGenerator | dict
    ) -> None:
        super().__init__()
        self._mean = setup_sequence_generator(mean)
        self._std = setup_sequence_generator(std)

    def __repr__(self) -> str:
        args = str_indent(str_mapping({"mean": self._mean, "std": self._std}))
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def generate(
        self, seq_len: int, batch_size: int = 1, rng: Generator | None = None
    ) -> BatchedTensorSeq:
        return BatchedTensorSeq(
            normal(
                mean=self._mean.generate(seq_len=seq_len, batch_size=batch_size, rng=rng).data,
                std=self._std.generate(seq_len=seq_len, batch_size=batch_size, rng=rng).data,
                generator=rng,
            )
        )


class RandNormalSequenceGenerator(BaseSequenceGenerator):
    r"""Implements a sequence generator to generate cyclic sequences by
    sampling values from a Normal distribution.

    Args:
    ----
        mean (float, optional): Specifies the mean of the Normal
            distribution. Default: ``0.0``
        std (float, optional): Specifies the standard deviation of the
            Normal distribution. Default: ``1.0``
        feature_size (tuple or list or int, optional): Specifies the
            feature size. Default: ``1``

    Raises:
    ------
        ValueError if ``std`` is not a positive number.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.sequence import RandNormal
        >>> generator = RandNormal(mean=0.0, std=1.0)
        >>> generator
        RandNormalSequenceGenerator(mean=0.0, std=1.0, feature_size=(1,))
        >>> generator.generate(seq_len=6, batch_size=2)
        tensor([[...]], batch_dim=0, seq_dim=1)
    """

    def __init__(
        self,
        mean: float = 0.0,
        std: float = 1.0,
        feature_size: tuple[int, ...] | list[int] | int = 1,
    ) -> None:
        super().__init__()
        self._mean = float(mean)
        if std <= 0:
            raise ValueError(f"std has to be greater than 0 (received: {std})")
        self._std = float(std)
        self._feature_size = to_tuple(feature_size)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(mean={self._mean}, std={self._std}, "
            f"feature_size={self._feature_size})"
        )

    def generate(
        self, seq_len: int, batch_size: int = 1, rng: Generator | None = None
    ) -> BatchedTensorSeq:
        return BatchedTensorSeq(
            rand_normal(
                size=(batch_size, seq_len) + self._feature_size,
                mean=self._mean,
                std=self._std,
                generator=rng,
            )
        )


class RandTruncNormalSequenceGenerator(BaseSequenceGenerator):
    r"""Implements a sequence generator to generate cyclic sequences by
    sampling values from a truncated Normal distribution.

    Args:
    ----
        mean (float, optional): Specifies the mean of the Normal
            distribution. Default: ``0.0``
        std (float, optional): Specifies the standard deviation of
            the Normal distribution. Default: ``1.0``
        min_value (float, optional): Specifies the minimum value.
            Default: ``-3.0``
        max_value (float, optional): Specifies the maximum value.
            Default: ``3.0``
        feature_size (tuple or list or int, optional): Specifies the
            feature size. Default: ``1``

    Raises:
    ------
        ValueError if ``std`` is not a positive number.
        ValueError if ``max_value`` is lower than ``min_value``.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.sequence import RandTruncNormal
        >>> generator = RandTruncNormal(mean=0.0, std=1.0, min_value=-1.0, max_value=1.0)
        >>> generator
        RandTruncNormalSequenceGenerator(mean=0.0, std=1.0, min_value=-1.0, max_value=1.0, feature_size=(1,))
        >>> generator.generate(seq_len=6, batch_size=2)
        tensor([[...]], batch_dim=0, seq_dim=1)
    """

    def __init__(
        self,
        mean: float = 0.0,
        std: float = 1.0,
        min_value: float = -3.0,
        max_value: float = 3.0,
        feature_size: tuple[int, ...] | list[int] | int = 1,
    ) -> None:
        super().__init__()
        self._mean = float(mean)
        if std <= 0:
            raise ValueError(f"std has to be greater than 0 (received: {std})")
        self._std = float(std)
        if max_value < min_value:
            raise ValueError(
                f"max_value ({max_value}) has to be greater or equal to min_value ({min_value})"
            )
        self._min_value = float(min_value)
        self._max_value = float(max_value)
        self._feature_size = to_tuple(feature_size)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(mean={self._mean}, std={self._std}, "
            f"min_value={self._min_value}, max_value={self._max_value}, "
            f"feature_size={self._feature_size})"
        )

    def generate(
        self, seq_len: int, batch_size: int = 1, rng: Generator | None = None
    ) -> BatchedTensorSeq:
        return BatchedTensorSeq(
            rand_trunc_normal(
                size=(batch_size, seq_len) + self._feature_size,
                mean=self._mean,
                std=self._std,
                min_value=self._min_value,
                max_value=self._max_value,
                generator=rng,
            )
        )


class TruncNormalSequenceGenerator(BaseSequenceGenerator):
    r"""Implements a class to generate sequence by sampling values from a
    truncated Normal distribution.

    Args:
    ----
        mean (``BaseSequenceGenerator`` or dict): Specifies a sequence
            generator (or its configuration) to generate the mean.
        std (``BaseSequenceGenerator`` or dict): Specifies a sequence
            generator (or its configuration) to generate the standard
            deviation.
        min_value (``BaseSequenceGenerator`` or dict): Specifies a
            sequence generator (or its configuration) to generate the
            minimum value (included).
        max_value (``BaseSequenceGenerator`` or dict): Specifies a
            sequence generator (or its configuration) to generate the
            maximum value (excluded).

    Example usage:

    .. code-block:: pycon

        >>> from startorch.sequence import RandUniform, TruncNormal
        >>> generator = TruncNormal(
        ...     mean=RandUniform(low=-1.0, high=1.0),
        ...     std=RandUniform(low=1.0, high=2.0),
        ...     min_value=RandUniform(low=-10.0, high=-5.0),
        ...     max_value=RandUniform(low=5.0, high=10.0),
        ... )
        >>> generator
        TruncNormalSequenceGenerator(
          (mean): RandUniformSequenceGenerator(low=-1.0, high=1.0, feature_size=(1,))
          (std): RandUniformSequenceGenerator(low=1.0, high=2.0, feature_size=(1,))
          (min_value): RandUniformSequenceGenerator(low=-10.0, high=-5.0, feature_size=(1,))
          (max_value): RandUniformSequenceGenerator(low=5.0, high=10.0, feature_size=(1,))
        )
        >>> generator.generate(seq_len=6, batch_size=2)
        tensor([[...]], batch_dim=0, seq_dim=1)
    """

    def __init__(
        self,
        mean: BaseSequenceGenerator | dict,
        std: BaseSequenceGenerator | dict,
        min_value: BaseSequenceGenerator | dict,
        max_value: BaseSequenceGenerator | dict,
    ) -> None:
        super().__init__()
        self._mean = setup_sequence_generator(mean)
        self._std = setup_sequence_generator(std)
        self._min_value = setup_sequence_generator(min_value)
        self._max_value = setup_sequence_generator(max_value)

    def __repr__(self) -> str:
        args = str_indent(
            str_mapping(
                {
                    "mean": self._mean,
                    "std": self._std,
                    "min_value": self._min_value,
                    "max_value": self._max_value,
                }
            )
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def generate(
        self, seq_len: int, batch_size: int = 1, rng: Generator | None = None
    ) -> BatchedTensorSeq:
        return BatchedTensorSeq(
            trunc_normal(
                mean=self._mean.generate(seq_len=seq_len, batch_size=batch_size, rng=rng).data,
                std=self._std.generate(seq_len=seq_len, batch_size=batch_size, rng=rng).data,
                min_value=self._min_value.generate(
                    seq_len=seq_len, batch_size=batch_size, rng=rng
                ).data,
                max_value=self._max_value.generate(
                    seq_len=seq_len, batch_size=batch_size, rng=rng
                ).data,
                generator=rng,
            )
        )



from __future__ import annotations

__all__ = [
    "exponential",
    "half_cauchy",
    "half_normal",
    "log_normal",
    "rand_exponential",
    "rand_half_cauchy",
    "rand_half_normal",
    "rand_log_normal",
]


import torch
from torch import Generator, Tensor

from startorch.random import cauchy, normal, rand_cauchy, rand_normal


def rand_exponential(
    size: list[int] | tuple[int, ...],
    rate: float = 1.0,
    generator: Generator | None = None,
) -> Tensor:
    r"""Creates a tensor filled with values sampled from an Exponential
    distribution.

    Args:
    ----
        size (list or tuple): Specifies the tensor shape.
        rate (float, optional): Specifies the rate of the Exponential
            distribution. Default: ``1.0``
        generator (``torch.Generator`` or None, optional): Specifies
            an optional random generator. Default: ``None``

    Returns:
    -------
        ``torch.Tensor``: A tensor filled with values sampled from an
            Exponential distribution

    Raises:
    ------
        ValueError if the ``rate`` parameter is not valid.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from startorch.random import rand_exponential
        >>> rand_exponential((2, 3), rate=1.0)
        tensor([[...]])
    """
    if rate <= 0:
        raise ValueError(f"rate has to be greater than 0 (received: {rate})")
    tensor = torch.zeros(*size, dtype=torch.float)
    tensor.exponential_(rate, generator=generator)
    return tensor


def exponential(rate: torch.Tensor, generator: torch.Generator | None = None) -> torch.Tensor:
    r"""Creates a tensor filled with values sampled from an Exponential
    distribution.

    Unlike ``rand_exponential``, this function allows to sample values
    from different Exponential distributions at the same time.
    The shape of the ``rate`` tensor is used to infer the output size.

    Args:
    ----
        rate (``torch.Tensor`` of type float and shape
            ``(d0, d1, ..., dn)``): Specifies the rates
            of the Exponential distribution. The output has the
            same shape that this input.
        generator (``torch.Generator`` or None, optional): Specifies
            an optional random generator. Default: ``None``

    Returns:
    -------
        ``torch.Tensor`` of type float and shape
            ``(d0, d1, ..., dn)``: A tensor filled with values sampled
            from an Exponential distribution.

    Raises:
    ------
        ValueError if the ``rate`` parameter is not valid.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from startorch.random import exponential
        >>> exponential(torch.tensor([1.0, 3.0, 5.0]))
        tensor([...])
    """
    rate = rate.float()
    if torch.any(rate <= 0.0):
        raise ValueError("rate values have to be greater than 0 (>0)")
    return torch.zeros_like(rate).exponential_(generator=generator) / rate


def rand_half_cauchy(
    size: list[int] | tuple[int, ...],
    scale: float = 1.0,
    generator: Generator | None = None,
) -> Tensor:
    r"""Creates a tensor filled with values sampled from a half-Cauchy
    distribution.

    Args:
    ----
        size (list or tuple): Specifies the tensor shape.
        scale (float, optional): Specifies the scale of the
            half-Cauchy distribution. This value has to be greater
            than 0. Default: ``1.0``
        generator (``torch.Generator`` or None, optional): Specifies
            an optional random generator. Default: ``None``

    Returns:
    -------
        ``torch.Tensor``: A tensor filled with values sampled from a
            half-Cauchy distribution.

    Raises:
    ------
        ValueError if the ``scale`` parameter is not valid.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from startorch.random import rand_half_cauchy
        >>> rand_half_cauchy((2, 3), scale=1.0)
        tensor([[...]])
    """
    return rand_cauchy(size=size, loc=0.0, scale=scale, generator=generator).abs()


def half_cauchy(scale: Tensor, generator: Generator | None = None) -> Tensor:
    r"""Creates a tensor filled with values sampled from a half-Cauchy
    distribution.

    Unlike ``rand_half_cauchy``, this function allows to sample values
    from different half-Cauchy distributions at the same time.
    The shape of the ``scale`` tensor is used to infer
    the output size.

    Args:
    ----
        scale (``torch.Tensor`` of type float and shape
            ``(d0, d1, ..., dn)``): Specifies the scale
            of the half-Cauchy distribution.
        generator (``torch.Generator`` or None, optional): Specifies
            an optional random generator. Default: ``None``

    Returns:
    -------
        ``torch.Tensor``: A tensor filled with values sampled from a
            half-Cauchy distribution.

    Raises:
    ------
        ValueError if the ``scale`` parameter is not valid.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from startorch.random import half_cauchy
        >>> half_cauchy(torch.tensor([1.0, 3.0, 5.0]))
        tensor([...])
    """
    return cauchy(loc=torch.zeros_like(scale), scale=scale, generator=generator).abs()


def rand_half_normal(
    size: list[int] | tuple[int, ...],
    std: float = 1.0,
    generator: Generator | None = None,
) -> Tensor:
    r"""Creates a tensor filled with values sampled from a half-Normal
    distribution.

    Args:
    ----
        size (list or tuple): Specifies the tensor shape.
        std (float, optional): Specifies the standard deviation of
            the half-Normal distribution. Default: ``1.0``
        generator (``torch.Generator`` or None, optional): Specifies
            an optional random generator. Default: ``None``

    Returns:
    -------
        ``torch.Tensor``: A tensor filled with values sampled from a
            half-Normal distribution.

    Raises:
    ------
        ValueError if the ``std`` parameter is not valid.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from startorch.random import rand_half_normal
        >>> rand_half_normal((2, 3), std=1.0)
        tensor([[...]])
    """
    if std <= 0:
        raise ValueError(f"std has to be greater than 0 (received: {std})")
    return rand_normal(size=size, mean=0.0, std=std, generator=generator).abs()


def half_normal(std: Tensor, generator: Generator | None = None) -> Tensor:
    r"""Creates a tensor filled with values sampled from a half-Normal
    distribution.

    Unlike ``rand_half_normal``, this function allows to sample values
    from different half-Normal distributions at the same time.
    The shape of the ``std`` tensor is used to infer
    the output size.

    Args:
    ----
        std (``torch.Tensor`` of type float and shape
            ``(d0, d1, ..., dn)``): Specifies the standard
            deviation of the half-Normal distribution.
        generator (``torch.Generator`` or None, optional): Specifies
            an optional random generator. Default: ``None``

    Returns:
    -------
        ``torch.Tensor``: A tensor filled with values sampled from a
            half-Normal distribution.

    Raises:
    ------
        ValueError if the ``std`` parameter is not valid.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from startorch.random import half_normal
        >>> half_normal(torch.tensor([1.0, 3.0, 5.0]))
        tensor([...])
    """
    return normal(mean=torch.zeros_like(std), std=std, generator=generator).abs()


def rand_log_normal(
    size: list[int] | tuple[int, ...],
    mean: float = 0.0,
    std: float = 1.0,
    generator: Generator | None = None,
) -> Tensor:
    r"""Creates a tensor filled with values sampled from a log-Normal
    distribution.

    Args:
    ----
        size (list or tuple): Specifies the tensor shape.
        mean (float, optional): Specifies the mean of the underlying
            Normal distribution. Default: ``0.0``
        std (float, optional): Specifies the standard deviation of
            the underlying Normal distribution. Default: ``1.0``
        generator (``torch.Generator`` or None, optional): Specifies
            an optional random generator. Default: ``None``

    Returns:
    -------
        ``torch.Tensor``: A tensor filled with values sampled from a
            log-Normal distribution.

    Raises:
    ------
        ValueError if the ``mean`` and ``std`` parametera are not
            valid.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from startorch.random import rand_log_normal
        >>> rand_log_normal((2, 3), mean=1.0, std=2.0)
        tensor([[...]])
    """
    if std <= 0:
        raise ValueError(f"std has to be greater than 0 (received: {std})")
    tensor = torch.zeros(*size, dtype=torch.float)
    tensor.log_normal_(mean=mean, std=std, generator=generator)
    return tensor


def log_normal(mean: Tensor, std: Tensor, generator: Generator | None = None) -> Tensor:
    r"""Creates a tensor filled with values sampled from a log-Normal
    distribution.

    Unlike ``rand_log_normal``, this function allows to sample values
    from different log-Normal distributions at the same time.
    The shape of the ``mean`` and ``std`` tensors are used to infer
    the output size.

    Args:
    ----
        mean (``torch.Tensor`` of type float and shape
            ``(d0, d1, ..., dn)``): Specifies the mean
            of the log-Normal distribution.
        std (``torch.Tensor`` of type float and shape
            ``(d0, d1, ..., dn)``): Specifies the standard
            deviation of the log-Normal distribution.
        generator (``torch.Generator`` or None, optional): Specifies
            an optional random generator. Default: ``None``

    Returns:
    -------
        ``torch.Tensor``: A tensor filled with values sampled from a
            log-Normal distribution.

    Raises:
    ------
        ValueError if the ``mean`` and ``std`` parametera are not
            valid.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from startorch.random import log_normal
        >>> log_normal(torch.tensor([-1.0, 0.0, 1.0]), torch.tensor([1.0, 3.0, 5.0]))
        tensor([...])
    """
    return normal(mean=mean, std=std, generator=generator).exp()



from __future__ import annotations

__all__ = [
    "CauchyTensorGenerator",
    "RandCauchyTensorGenerator",
    "RandTruncCauchyTensorGenerator",
    "TruncCauchyTensorGenerator",
]


from coola.utils.format import str_indent, str_mapping
from torch import Generator, Tensor

from startorch.random import cauchy, rand_cauchy, rand_trunc_cauchy, trunc_cauchy
from startorch.tensor.base import BaseTensorGenerator, setup_tensor_generator


class CauchyTensorGenerator(BaseTensorGenerator):
    r"""Implements a class to generate tensor by sampling values from a
    Cauchy distribution.

    Args:
    ----
        loc (``BaseTensorGenerator`` or dict): Specifies a tensor
            generator (or its configuration) to generate the location.
        scale (``BaseTensorGenerator`` or dict): Specifies a tensor
            generator (or its configuration) to generate the scale.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.tensor import Cauchy, RandUniform
        >>> generator = Cauchy(
        ...     loc=RandUniform(low=-1.0, high=1.0), scale=RandUniform(low=1.0, high=2.0)
        ... )
        >>> generator
        CauchyTensorGenerator(
          (loc): RandUniformTensorGenerator(low=-1.0, high=1.0)
          (scale): RandUniformTensorGenerator(low=1.0, high=2.0)
        )
        >>> generator.generate((2, 6))
        tensor([[...]])
    """

    def __init__(
        self,
        loc: BaseTensorGenerator | dict,
        scale: BaseTensorGenerator | dict,
    ) -> None:
        super().__init__()
        self._loc = setup_tensor_generator(loc)
        self._scale = setup_tensor_generator(scale)

    def __repr__(self) -> str:
        args = str_indent(str_mapping({"loc": self._loc, "scale": self._scale}))
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def generate(self, size: tuple[int, ...], rng: Generator | None = None) -> Tensor:
        return cauchy(
            loc=self._loc.generate(size=size, rng=rng).data,
            scale=self._scale.generate(size=size, rng=rng).data,
            generator=rng,
        )


class RandCauchyTensorGenerator(BaseTensorGenerator):
    r"""Implements a class to generate tensor by sampling values from a
    Cauchy distribution.

    Args:
    ----
        loc (float, optional): Specifies the location/median of the
            Cauchy distribution. Default: ``0.0``
        scale (float, optional): Specifies the scale of the
            distribution. Default: ``1.0``

    Raises:
    ------
        ValueError if ``scale`` is not a positive number.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.tensor import RandCauchy
        >>> generator = RandCauchy(loc=0.0, scale=1.0)
        >>> generator
        RandCauchyTensorGenerator(loc=0.0, scale=1.0)
        >>> generator.generate((2, 6))
        tensor([[...]])
    """

    def __init__(
        self,
        loc: float = 0.0,
        scale: float = 1.0,
    ) -> None:
        super().__init__()
        self._loc = float(loc)
        if scale <= 0:
            raise ValueError(f"scale has to be greater than 0 (received: {scale})")
        self._scale = float(scale)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(loc={self._loc}, scale={self._scale})"

    def generate(self, size: tuple[int, ...], rng: Generator | None = None) -> Tensor:
        return rand_cauchy(
            size=size,
            loc=self._loc,
            scale=self._scale,
            generator=rng,
        )


class RandTruncCauchyTensorGenerator(BaseTensorGenerator):
    r"""Implements a class to generate tensor by sampling values from a
    truncated Cauchy distribution.

    Args:
    ----
        loc (float, optional): Specifies the location/median of the
            Cauchy distribution. Default: ``0.0``
        scale (float, optional): Specifies the scale of the
            distribution. Default: ``1.0``
        min_value (float, optional): Specifies the minimum value.
            (included). Default: ``-2.0``
        max_value (float, optional): Specifies the maximum value
            (excluded). Default: ``2.0``

    Raises:
    ------
        ValueError if ``std`` is not a positive number.
        ValueError if ``max_value`` is lower than ``min_value``.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.tensor import RandTruncCauchy
        >>> generator = RandTruncCauchy(loc=0.0, scale=1.0, min_value=-1.0, max_value=1.0)
        >>> generator
        RandTruncCauchyTensorGenerator(loc=0.0, scale=1.0, min_value=-1.0, max_value=1.0)
        >>> generator.generate((2, 6))
        tensor([[...]])
    """

    def __init__(
        self,
        loc: float = 0.0,
        scale: float = 1.0,
        min_value: float = -2.0,
        max_value: float = 2.0,
    ) -> None:
        super().__init__()
        self._loc = float(loc)
        if scale <= 0:
            raise ValueError(f"scale has to be greater than 0 (received: {scale})")
        self._scale = float(scale)
        if max_value < min_value:
            raise ValueError(
                f"max_value ({max_value}) has to be greater or equal to min_value ({min_value})"
            )
        self._min_value = float(min_value)
        self._max_value = float(max_value)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(loc={self._loc}, scale={self._scale}, "
            f"min_value={self._min_value}, max_value={self._max_value})"
        )

    def generate(self, size: tuple[int, ...], rng: Generator | None = None) -> Tensor:
        return rand_trunc_cauchy(
            size=size,
            loc=self._loc,
            scale=self._scale,
            min_value=self._min_value,
            max_value=self._max_value,
            generator=rng,
        )


class TruncCauchyTensorGenerator(BaseTensorGenerator):
    r"""Implements a class to generate tensor by sampling values from a
    Cauchy distribution.

    Args:
    ----
        loc (``BaseTensorGenerator`` or dict): Specifies a tensor
            generator (or its configuration) to generate the location.
        scale (``BaseTensorGenerator`` or dict): Specifies a tensor
            generator (or its configuration) to generate the scale.
        min_value (``BaseTensorGenerator`` or dict): Specifies a
            tensor generator (or its configuration) to generate the
            minimum value (included).
        max_value (``BaseTensorGenerator`` or dict): Specifies a
            tensor generator (or its configuration) to generate the
            maximum value (excluded).

    Example usage:

    .. code-block:: pycon

        >>> from startorch.tensor import RandUniform, TruncCauchy
        >>> generator = TruncCauchy(
        ...     loc=RandUniform(low=-1.0, high=1.0),
        ...     scale=RandUniform(low=1.0, high=2.0),
        ...     min_value=RandUniform(low=-10.0, high=-5.0),
        ...     max_value=RandUniform(low=5.0, high=10.0),
        ... )
        >>> generator
        TruncCauchyTensorGenerator(
          (loc): RandUniformTensorGenerator(low=-1.0, high=1.0)
          (scale): RandUniformTensorGenerator(low=1.0, high=2.0)
          (min_value): RandUniformTensorGenerator(low=-10.0, high=-5.0)
          (max_value): RandUniformTensorGenerator(low=5.0, high=10.0)
        )
        >>> generator.generate((2, 6))
        tensor([[...]])
    """

    def __init__(
        self,
        loc: BaseTensorGenerator | dict,
        scale: BaseTensorGenerator | dict,
        min_value: BaseTensorGenerator | dict,
        max_value: BaseTensorGenerator | dict,
    ) -> None:
        super().__init__()
        self._loc = setup_tensor_generator(loc)
        self._scale = setup_tensor_generator(scale)
        self._min_value = setup_tensor_generator(min_value)
        self._max_value = setup_tensor_generator(max_value)

    def __repr__(self) -> str:
        args = str_indent(
            str_mapping(
                {
                    "loc": self._loc,
                    "scale": self._scale,
                    "min_value": self._min_value,
                    "max_value": self._max_value,
                }
            )
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def generate(self, size: tuple[int, ...], rng: Generator | None = None) -> Tensor:
        return trunc_cauchy(
            loc=self._loc.generate(size=size, rng=rng),
            scale=self._scale.generate(size=size, rng=rng),
            min_value=self._min_value.generate(size=size, rng=rng),
            max_value=self._max_value.generate(size=size, rng=rng),
            generator=rng,
        )



from __future__ import annotations

__all__ = ["TensorSequenceGenerator"]


from coola.utils.format import str_indent, str_mapping
from redcat import BatchedTensorSeq
from torch import Generator

from startorch.sequence.base import BaseSequenceGenerator
from startorch.tensor.base import BaseTensorGenerator, setup_tensor_generator
from startorch.utils.conversion import to_tuple


class TensorSequenceGenerator(BaseSequenceGenerator):
    r"""Implements a sequence generator to generate sequences from a
    tensor generator.

    Args:
    ----
        tensor (``BaseTensorGenerator`` or dict): Specifies a tensor
            generator (or its configuration).
        feature_size (tuple or list or int, optional): Specifies the
            feature size. Default: ``1``

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from startorch.sequence import TensorSequence
        >>> from startorch.tensor import RandUniform
        >>> generator = TensorSequence(RandUniform())
        >>> generator
        TensorSequenceGenerator(
          (tensor): RandUniformTensorGenerator(low=0.0, high=1.0)
          (feature_size): ()
        )
        >>> generator.generate(seq_len=12, batch_size=4)
        tensor([[...]], batch_dim=0, seq_dim=1)
    """

    def __init__(
        self,
        tensor: BaseTensorGenerator | dict,
        feature_size: tuple[int, ...] | list[int] | int = tuple(),
    ) -> None:
        super().__init__()
        self._tensor = setup_tensor_generator(tensor)
        self._feature_size = to_tuple(feature_size)

    def __repr__(self) -> str:
        args = str_indent(str_mapping({"tensor": self._tensor, "feature_size": self._feature_size}))
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def generate(
        self, seq_len: int, batch_size: int = 1, rng: Generator | None = None
    ) -> BatchedTensorSeq:
        return BatchedTensorSeq(
            self._tensor.generate(size=(batch_size, seq_len) + self._feature_size, rng=rng)
        )



from __future__ import annotations

__all__ = ["PeriodicTimeSeriesGenerator"]

import math

from coola.utils import str_indent, str_mapping
from redcat import BatchDict
from torch import Generator

from startorch.periodic.timeseries.base import BasePeriodicTimeSeriesGenerator
from startorch.tensor.base import BaseTensorGenerator, setup_tensor_generator
from startorch.timeseries.base import (
    BaseTimeSeriesGenerator,
    setup_timeseries_generator,
)


class PeriodicTimeSeriesGenerator(BaseTimeSeriesGenerator):
    r"""Implements a time series generator to generate periodic time
    series from a regular time series generator.

    Args:
    ----
        timeseries (``BaseTimeSeriesGenerator`` or
            ``BasePeriodicTimeSeriesGenerator`` or dict): Specifies a
            time series generator or its configuration that is used to
            generate the periodic pattern.
        period (``BaseTensorGenerator`` or dict): Specifies
            the period length sampler or its configuration. This
            sampler is used to sample the period length at each
            batch.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.timeseries import Periodic, TimeSeries
        >>> from startorch.sequence import RandUniform
        >>> from startorch.tensor import RandInt
        >>> generator = Periodic(
        ...     TimeSeries({"value": RandUniform(), "time": RandUniform()}), period=RandInt(2, 5)
        ... )
        >>> generator
        PeriodicTimeSeriesGenerator(
          (sequence): TimeSeriesGenerator(
              (value): RandUniformSequenceGenerator(low=0.0, high=1.0, feature_size=(1,))
              (time): RandUniformSequenceGenerator(low=0.0, high=1.0, feature_size=(1,))
            )
          (period): RandIntTensorGenerator(low=2, high=5)
        )
        >>> generator.generate(seq_len=10, batch_size=2)
        BatchDict(
          (value): tensor([[...]], batch_dim=0, seq_dim=1)
          (time): tensor([[...]], batch_dim=0, seq_dim=1)
        )
    """

    def __init__(
        self,
        timeseries: BaseTimeSeriesGenerator | BasePeriodicTimeSeriesGenerator | dict,
        period: BaseTensorGenerator | dict,
    ) -> None:
        super().__init__()
        self._timeseries = setup_timeseries_generator(timeseries)
        self._period = setup_tensor_generator(period)

    def __repr__(self) -> str:
        args = str_indent(str_mapping({"sequence": self._timeseries, "period": self._period}))
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def generate(
        self, seq_len: int, batch_size: int = 1, rng: Generator | None = None
    ) -> BatchDict:
        period = int(self._period.generate((1,), rng=rng).item())
        if isinstance(self._timeseries, BasePeriodicTimeSeriesGenerator):
            return self._timeseries.generate(
                seq_len=seq_len, period=period, batch_size=batch_size, rng=rng
            )
        return (
            self._timeseries.generate(seq_len=period, batch_size=batch_size, rng=rng)
            .repeat_along_seq(math.ceil(seq_len / period))
            .slice_along_seq(stop=seq_len)
        )



from __future__ import annotations

__all__ = [
    "AsinhUniformSequenceGenerator",
    "LogUniformSequenceGenerator",
    "RandAsinhUniformSequenceGenerator",
    "RandIntSequenceGenerator",
    "RandLogUniformSequenceGenerator",
    "RandUniformSequenceGenerator",
    "UniformSequenceGenerator",
]

from collections.abc import Generator

import torch
from coola.utils import str_indent, str_mapping
from redcat import BatchedTensorSeq

from startorch.random import (
    asinh_uniform,
    log_uniform,
    rand_asinh_uniform,
    rand_log_uniform,
    rand_uniform,
    uniform,
)
from startorch.sequence.base import BaseSequenceGenerator, setup_sequence_generator
from startorch.utils.conversion import to_tuple


class AsinhUniformSequenceGenerator(BaseSequenceGenerator):
    r"""Implements a sequence generator to generate sequences by sampling
    values from an asinh-uniform distribution.

    Args:
    ----
        low (``BaseSequenceGenerator`` or dict): Specifies a sequence
            generator (or its configuration) to generate the minimum
            value (inclusive).
        high (``BaseSequenceGenerator`` or dict): Specifies a sequence
            generator (or its configuration) to generate the maximum
            value (exclusive).

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from startorch.sequence import AsinhUniform, RandUniform
        >>> generator = AsinhUniform(low=RandUniform(-1.0, 0.0), high=RandUniform(0.0, 1.0))
        >>> generator
        AsinhUniformSequenceGenerator(
          (low): RandUniformSequenceGenerator(low=-1.0, high=0.0, feature_size=(1,))
          (high): RandUniformSequenceGenerator(low=0.0, high=1.0, feature_size=(1,))
        )
        >>> generator.generate(seq_len=12, batch_size=4)
        tensor([[...]], batch_dim=0, seq_dim=1)
    """

    def __init__(
        self,
        low: BaseSequenceGenerator | dict,
        high: BaseSequenceGenerator | dict,
    ) -> None:
        super().__init__()
        self._low = setup_sequence_generator(low)
        self._high = setup_sequence_generator(high)

    def __repr__(self) -> str:
        args = str_indent(str_mapping({"low": self._low, "high": self._high}))
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def generate(
        self, seq_len: int, batch_size: int = 1, rng: Generator | None = None
    ) -> BatchedTensorSeq:
        return BatchedTensorSeq(
            asinh_uniform(
                low=self._low.generate(seq_len=seq_len, batch_size=batch_size, rng=rng).data,
                high=self._high.generate(seq_len=seq_len, batch_size=batch_size, rng=rng).data,
                generator=rng,
            )
        )


class LogUniformSequenceGenerator(BaseSequenceGenerator):
    r"""Implements a sequence generator to generate sequences by sampling
    values from a log-uniform distribution.

    Args:
    ----
        low (``BaseSequenceGenerator`` or dict): Specifies a sequence
            generator (or its configuration) to generate the minimum
            value (inclusive).
        high (``BaseSequenceGenerator`` or dict): Specifies a sequence
            generator (or its configuration) to generate the maximum
            value (exclusive).

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from startorch.sequence import LogUniform, RandUniform
        >>> generator = LogUniform(low=RandUniform(0.0, 1.0), high=RandUniform(5.0, 10.0))
        >>> generator
        LogUniformSequenceGenerator(
          (low): RandUniformSequenceGenerator(low=0.0, high=1.0, feature_size=(1,))
          (high): RandUniformSequenceGenerator(low=5.0, high=10.0, feature_size=(1,))
        )
        >>> generator.generate(seq_len=12, batch_size=4)
        tensor([[...]], batch_dim=0, seq_dim=1)
    """

    def __init__(
        self,
        low: BaseSequenceGenerator | dict,
        high: BaseSequenceGenerator | dict,
    ) -> None:
        super().__init__()
        self._low = setup_sequence_generator(low)
        self._high = setup_sequence_generator(high)

    def __repr__(self) -> str:
        args = str_indent(str_mapping({"low": self._low, "high": self._high}))
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def generate(
        self, seq_len: int, batch_size: int = 1, rng: Generator | None = None
    ) -> BatchedTensorSeq:
        return BatchedTensorSeq(
            log_uniform(
                low=self._low.generate(seq_len=seq_len, batch_size=batch_size, rng=rng).data,
                high=self._high.generate(seq_len=seq_len, batch_size=batch_size, rng=rng).data,
                generator=rng,
            )
        )


class RandAsinhUniformSequenceGenerator(BaseSequenceGenerator):
    r"""Implements a sequence generator to generate sequences by sampling
    values from an asinh-uniform distribution.

    Args:
    ----
        low (float): Specifies the minimum value (inclusive).
        high (float): Specifies the maximum value (exclusive).
        feature_size (tuple or list or int, optional): Specifies the
            feature size. Default: ``1``

    Raises:
    ------
        ValueError if ``high`` is lower than ``low``.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from startorch.sequence import RandAsinhUniform
        >>> generator = RandAsinhUniform(low=1.0, high=10.0)
        >>> generator
        RandAsinhUniformSequenceGenerator(low=1.0, high=10.0, feature_size=(1,))
        >>> generator.generate(seq_len=12, batch_size=4)
        tensor([[...]], batch_dim=0, seq_dim=1)
    """

    def __init__(
        self,
        low: float,
        high: float,
        feature_size: tuple[int, ...] | list[int] | int = 1,
    ) -> None:
        super().__init__()
        self._low = float(low)
        if high < low:
            raise ValueError(f"high ({high}) has to be greater or equal to low ({low})")
        self._high = float(high)
        self._feature_size = to_tuple(feature_size)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(low={self._low}, "
            f"high={self._high}, feature_size={self._feature_size})"
        )

    def generate(
        self, seq_len: int, batch_size: int = 1, rng: Generator | None = None
    ) -> BatchedTensorSeq:
        return BatchedTensorSeq(
            rand_asinh_uniform(
                size=(batch_size, seq_len) + self._feature_size,
                low=self._low,
                high=self._high,
                generator=rng,
            )
        )


class RandIntSequenceGenerator(BaseSequenceGenerator):
    r"""Implements a class to generate sequences of uniformly distributed
    integers.

    Args:
    ----
        low (int): Specifies the minimum value (included).
        high (int): Specifies the maximum value (excluded).
        feature_size (tuple or list or int, optional): Specifies the
            feature size. Default: ``tuple()``

    Raises:
    ------
        ValueError if ``high`` is lower than ``low``.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from startorch.sequence import RandInt
        >>> generator = RandInt(0, 100)
        >>> generator
        RandIntSequenceGenerator(low=0, high=100, feature_size=())
        >>> generator.generate(seq_len=12, batch_size=4)
        tensor([[...]], batch_dim=0, seq_dim=1)
    """

    def __init__(
        self,
        low: int,
        high: int,
        feature_size: tuple[int, ...] | list[int] | int = tuple(),
    ) -> None:
        super().__init__()
        if high < low:
            raise ValueError(f"high ({high}) has to be greater or equal to low ({low})")
        self._low = int(low)
        self._high = int(high)
        self._feature_size = to_tuple(feature_size)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(low={self._low}, "
            f"high={self._high}, feature_size={self._feature_size})"
        )

    def generate(
        self, seq_len: int, batch_size: int = 1, rng: Generator | None = None
    ) -> BatchedTensorSeq:
        return BatchedTensorSeq(
            torch.randint(
                low=self._low,
                high=self._high,
                size=(batch_size, seq_len) + self._feature_size,
                generator=rng,
            )
        )


class RandLogUniformSequenceGenerator(BaseSequenceGenerator):
    r"""Implements a sequence generator to generate sequences by sampling
    values from a log-uniform distribution.

    Args:
    ----
        low (float): Specifies the minimum value (inclusive).
        high (float): Specifies the maximum value (exclusive).
        feature_size (tuple or list or int, optional): Specifies the
            feature size. Default: ``1``

    Raises:
    ------
        ValueError if ``high`` is lower than ``low``.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from startorch.sequence import RandLogUniform
        >>> generator = RandLogUniform(low=1.0, high=10.0)
        >>> generator
        RandLogUniformSequenceGenerator(low=1.0, high=10.0, feature_size=(1,))
        >>> generator.generate(seq_len=12, batch_size=4)
        tensor([[...]], batch_dim=0, seq_dim=1)
    """

    def __init__(
        self,
        low: float,
        high: float,
        feature_size: tuple[int, ...] | list[int] | int = 1,
    ) -> None:
        super().__init__()
        self._low = float(low)
        if high < low:
            raise ValueError(f"high ({high}) has to be greater or equal to low ({low})")
        self._high = float(high)
        self._feature_size = to_tuple(feature_size)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(low={self._low}, "
            f"high={self._high}, feature_size={self._feature_size})"
        )

    def generate(
        self, seq_len: int, batch_size: int = 1, rng: Generator | None = None
    ) -> BatchedTensorSeq:
        return BatchedTensorSeq(
            rand_log_uniform(
                size=(batch_size, seq_len) + self._feature_size,
                low=self._low,
                high=self._high,
                generator=rng,
            )
        )


class RandUniformSequenceGenerator(BaseSequenceGenerator):
    r"""Implements a sequence generator to generate sequences by sampling
    values from a uniform distribution.

    Args:
    ----
        low (float, optional): Specifies the minimum value
            (inclusive). Default: ``0.0``
        high (float, optional): Specifies the maximum value
            (exclusive). Default: ``1.0``
        feature_size (tuple or list or int, optional): Specifies the
            feature size. Default: ``1``

    Raises:
    ------
        ValueError if ``high`` is lower than ``low``.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from startorch.sequence import RandUniform
        >>> generator = RandUniform()
        >>> generator
        RandUniformSequenceGenerator(low=0.0, high=1.0, feature_size=(1,))
        >>> generator.generate(seq_len=12, batch_size=4)
        tensor([[...]], batch_dim=0, seq_dim=1)
    """

    def __init__(
        self,
        low: float = 0.0,
        high: float = 1.0,
        feature_size: tuple[int, ...] | list[int] | int = 1,
    ) -> None:
        super().__init__()
        self._low = float(low)
        if high < low:
            raise ValueError(f"high ({high}) has to be greater or equal to low ({low})")
        self._high = float(high)
        self._feature_size = to_tuple(feature_size)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(low={self._low}, "
            f"high={self._high}, feature_size={self._feature_size})"
        )

    def generate(
        self, seq_len: int, batch_size: int = 1, rng: Generator | None = None
    ) -> BatchedTensorSeq:
        return BatchedTensorSeq(
            rand_uniform(
                size=(batch_size, seq_len) + self._feature_size,
                low=self._low,
                high=self._high,
                generator=rng,
            )
        )


class UniformSequenceGenerator(BaseSequenceGenerator):
    r"""Implements a sequence generator to generate sequences by sampling
    values from a uniform distribution.

    Args:
        low (``BaseSequenceGenerator`` or dict): Specifies a sequence
            generator (or its configuration) to generate the minimum
            value (inclusive).
        high (``BaseSequenceGenerator`` or dict): Specifies a sequence
            generator (or its configuration) to generate the maximum
            value (exclusive).

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from startorch.sequence import Uniform, RandUniform
        >>> generator = Uniform(low=RandUniform(-1.0, 0.0), high=RandUniform(0.0, 1.0))
        >>> generator
        UniformSequenceGenerator(
          (low): RandUniformSequenceGenerator(low=-1.0, high=0.0, feature_size=(1,))
          (high): RandUniformSequenceGenerator(low=0.0, high=1.0, feature_size=(1,))
        )
        >>> generator.generate(seq_len=12, batch_size=4)
        tensor([[...]], batch_dim=0, seq_dim=1)
    """

    def __init__(
        self, low: BaseSequenceGenerator | dict, high: BaseSequenceGenerator | dict
    ) -> None:
        super().__init__()
        self._low = setup_sequence_generator(low)
        self._high = setup_sequence_generator(high)

    def __repr__(self) -> str:
        args = str_indent(str_mapping({"low": self._low, "high": self._high}))
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def generate(
        self, seq_len: int, batch_size: int = 1, rng: Generator | None = None
    ) -> BatchedTensorSeq:
        return BatchedTensorSeq(
            uniform(
                low=self._low.generate(seq_len=seq_len, batch_size=batch_size, rng=rng).data,
                high=self._high.generate(seq_len=seq_len, batch_size=batch_size, rng=rng).data,
                generator=rng,
            )
        )



from __future__ import annotations

__all__ = [
    "BaseExampleGenerator",
    "is_example_generator_config",
    "setup_example_generator",
]

import logging
from abc import ABC, abstractmethod
from collections.abc import Generator
from typing import Generic, TypeVar

from objectory import AbstractFactory
from objectory.utils import is_object_config
from redcat import BatchDict

from startorch.utils.format import str_target_object

logger = logging.getLogger(__name__)


T = TypeVar("T")


class BaseExampleGenerator(Generic[T], ABC, metaclass=AbstractFactory):
    r"""Defines the base class to generate time series.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.example import Hypercube
        >>> generator = Hypercube(num_classes=5, feature_size=6)
        >>> generator
        HypercubeExampleGenerator(num_classes=5, feature_size=6, noise_std=0.2)
        >>> batch = generator.generate(batch_size=10)
        >>> batch
        BatchDict(
          (target): tensor([...], batch_dim=0)
          (feature): tensor([[...]], batch_dim=0)
        )
    """

    @abstractmethod
    def generate(self, batch_size: int = 1, rng: Generator | None = None) -> BatchDict[T]:
        r"""Generates a time series.

        Args:
        ----
            batch_size (int, optional): Specifies the batch size.
                Default: ``1``
            rng (``torch.Generator`` or None, optional): Specifies
                an optional random number generator. Default: ``None``

        Returns:
        -------
            ``BatchDict``: A batch of time series.

        Example usage:

        .. code-block:: pycon

            >>> from startorch.example import Hypercube
            >>> generator = Hypercube(num_classes=5, feature_size=6)
            >>> batch = generator.generate(batch_size=10)
            >>> batch
            BatchDict(
              (target): tensor([...], batch_dim=0)
              (feature): tensor([[...]], batch_dim=0)
            )
        """


def is_example_generator_config(config: dict) -> bool:
    r"""Indicates if the input configuration is a configuration for a
    ``BaseExampleGenerator``.

    This function only checks if the value of the key  ``_target_``
    is valid. It does not check the other values. If ``_target_``
    indicates a function, the returned type hint is used to check
    the class.

    Args:
    ----
        config (dict): Specifies the configuration to check.

    Returns:
    -------
        bool: ``True`` if the input configuration is a configuration
            for a ``BaseExampleGenerator`` object.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.example import is_example_generator_config
        >>> is_example_generator_config({"_target_": "startorch.example.Hypercube"})
        True
    """
    return is_object_config(config, BaseExampleGenerator)


def setup_example_generator(
    generator: BaseExampleGenerator | dict,
) -> BaseExampleGenerator:
    r"""Sets up a time series generator.

    The time series generator is instantiated from its configuration
    by using the ``BaseExampleGenerator`` factory function.

    Args:
    ----
        generator (``BaseExampleGenerator`` or dict): Specifies a time
            series generator or its configuration.

    Returns:
    -------
        ``BaseExampleGenerator``: A time series generator.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.example import setup_example_generator
        >>> generator = setup_example_generator({"_target_": "startorch.example.Hypercube"})
        >>> generator
        HypercubeExampleGenerator(num_classes=50, feature_size=64, noise_std=0.2)
    """
    if isinstance(generator, dict):
        logger.info(
            "Initializing an example generator from its configuration... "
            f"{str_target_object(generator)}"
        )
        generator = BaseExampleGenerator.factory(**generator)
    if not isinstance(generator, BaseExampleGenerator):
        logger.warning(f"generator is not a `BaseExampleGenerator` (received: {type(generator)})")
    return generator



from __future__ import annotations

__all__ = [
    "BaseTimeSeriesGenerator",
    "is_timeseries_generator_config",
    "setup_timeseries_generator",
]

import logging
from abc import ABC, abstractmethod
from collections.abc import Generator

from objectory import AbstractFactory
from objectory.utils import is_object_config
from redcat import BatchDict

from startorch.utils.format import str_target_object

logger = logging.getLogger(__name__)


class BaseTimeSeriesGenerator(ABC, metaclass=AbstractFactory):
    r"""Defines the base class to generate time series.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from startorch.sequence import RandUniform
        >>> from startorch.timeseries import TimeSeries
        >>> generator = TimeSeries({"value": RandUniform(), "time": RandUniform()})
        >>> generator
        TimeSeriesGenerator(
          (value): RandUniformSequenceGenerator(low=0.0, high=1.0, feature_size=(1,))
          (time): RandUniformSequenceGenerator(low=0.0, high=1.0, feature_size=(1,))
        )
        >>> generator.generate(seq_len=12, batch_size=4)
        BatchDict(
          (value): tensor([[...]], batch_dim=0, seq_dim=1)
          (time): tensor([[...]], batch_dim=0, seq_dim=1)
        )
    """

    @abstractmethod
    def generate(
        self, seq_len: int, batch_size: int = 1, rng: Generator | None = None
    ) -> BatchDict:
        r"""Generates a time series.

        Args:
        ----
            seq_len (int): Specifies the sequence length.
            batch_size (int, optional): Specifies the batch size.
                Default: ``1``
            rng (``torch.Generator`` or None, optional): Specifies
                an optional random number generator. Default: ``None``

        Returns:
        -------
            ``BatchDict``: A batch of time series.

        Example usage:

        .. code-block:: pycon

            >>> import torch
            >>> from startorch.sequence import RandUniform
            >>> from startorch.timeseries import TimeSeries
            >>> generator = TimeSeries({"value": RandUniform(), "time": RandUniform()})
            >>> generator.generate(seq_len=12, batch_size=4)
            BatchDict(
              (value): tensor([[...]], batch_dim=0, seq_dim=1)
              (time): tensor([[...]], batch_dim=0, seq_dim=1)
            )
        """


def is_timeseries_generator_config(config: dict) -> bool:
    r"""Indicates if the input configuration is a configuration for a
    ``BaseTimeSeriesGenerator``.

    This function only checks if the value of the key  ``_target_``
    is valid. It does not check the other values. If ``_target_``
    indicates a function, the returned type hint is used to check
    the class.

    Args:
    ----
        config (dict): Specifies the configuration to check.

    Returns:
    -------
        bool: ``True`` if the input configuration is a configuration
            for a ``BaseTimeSeriesGenerator`` object.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.timeseries import is_timeseries_generator_config
        >>> is_timeseries_generator_config(
        ...     {
        ...         "_target_": "startorch.timeseries.TimeSeries",
        ...         "sequences": {
        ...             "value": {"_target_": "startorch.sequence.RandUniform"},
        ...             "time": {"_target_": "startorch.sequence.RandUniform"},
        ...         },
        ...     }
        ... )
        True
    """
    return is_object_config(config, BaseTimeSeriesGenerator)


def setup_timeseries_generator(
    generator: BaseTimeSeriesGenerator | dict,
) -> BaseTimeSeriesGenerator:
    r"""Sets up a time series generator.

    The time series generator is instantiated from its configuration
    by using the ``BaseTimeSeriesGenerator`` factory function.

    Args:
    ----
        generator (``BaseTimeSeriesGenerator`` or dict): Specifies a time
            series generator or its configuration.

    Returns:
    -------
        ``BaseTimeSeriesGenerator``: A time series generator.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.timeseries import setup_timeseries_generator
        >>> setup_timeseries_generator(
        ...     {
        ...         "_target_": "startorch.timeseries.TimeSeries",
        ...         "sequences": {
        ...             "value": {"_target_": "startorch.sequence.RandUniform"},
        ...             "time": {"_target_": "startorch.sequence.RandUniform"},
        ...         },
        ...     }
        ... )
        TimeSeriesGenerator(
          (value): RandUniformSequenceGenerator(low=0.0, high=1.0, feature_size=(1,))
          (time): RandUniformSequenceGenerator(low=0.0, high=1.0, feature_size=(1,))
        )
    """
    if isinstance(generator, dict):
        logger.info(
            "Initializing a time-series generator from its configuration... "
            f"{str_target_object(generator)}"
        )
        generator = BaseTimeSeriesGenerator.factory(**generator)
    if not isinstance(generator, BaseTimeSeriesGenerator):
        logger.warning(
            f"generator is not a `BaseTimeSeriesGenerator` (received: {type(generator)})"
        )
    return generator



from __future__ import annotations

__all__ = ["ConstantSequenceGenerator", "FullSequenceGenerator"]


import torch
from redcat import BatchedTensorSeq
from torch import Generator

from startorch.sequence.base import BaseSequenceGenerator
from startorch.sequence.wrapper import BaseWrapperSequenceGenerator
from startorch.utils.conversion import to_tuple


class ConstantSequenceGenerator(BaseWrapperSequenceGenerator):
    r"""Implements a sequence generator to generate a batch of sequences
    with constant values where the values for each sequence are sampled
    from another sequence generator.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.sequence import Constant, RandUniform
        >>> generator = Constant(RandUniform())
        >>> generator
        ConstantSequenceGenerator(
          (sequence): RandUniformSequenceGenerator(low=0.0, high=1.0, feature_size=(1,))
        )
        >>> generator.generate(seq_len=6, batch_size=2)
        tensor([[...]], batch_dim=0, seq_dim=1)
    """

    def generate(
        self, seq_len: int, batch_size: int = 1, rng: Generator | None = None
    ) -> BatchedTensorSeq:
        return self._generator.generate(seq_len=1, batch_size=batch_size, rng=rng).repeat_along_seq(
            seq_len
        )


class FullSequenceGenerator(BaseSequenceGenerator):
    r"""Implements a sequence generator to generate sequences filled with
    a given value.

    This sequence generator is fully deterministic and the random
    seed has no effect.

    Args:
    ----
        value (float): Specifies the value.
        feature_size (tuple or list or int, optional): Specifies the
            feature size. Default: ``1``

    Example usage:

    .. code-block:: pycon

        >>> from startorch.sequence import Full
        >>> generator = Full(42.0)
        >>> generator
        FullSequenceGenerator(value=42.0, feature_size=(1,))
        >>> generator.generate(seq_len=6, batch_size=2)
        tensor([[[42.],
                 [42.],
                 [42.],
                 [42.],
                 [42.],
                 [42.]],
                [[42.],
                 [42.],
                 [42.],
                 [42.],
                 [42.],
                 [42.]]], batch_dim=0, seq_dim=1)
    """

    def __init__(
        self,
        value: float,
        feature_size: tuple[int, ...] | list[int] | int = 1,
    ) -> None:
        super().__init__()
        self._value = float(value)
        self._feature_size = to_tuple(feature_size)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(value={self._value}, "
            f"feature_size={self._feature_size})"
        )

    def generate(
        self, seq_len: int, batch_size: int = 1, rng: Generator | None = None
    ) -> BatchedTensorSeq:
        return BatchedTensorSeq(torch.full((batch_size, seq_len) + self._feature_size, self._value))



from __future__ import annotations

__all__ = ["FloatTensorGenerator", "LongTensorGenerator"]

from torch import Generator, Tensor

from startorch.tensor.wrapper import BaseWrapperTensorGenerator


class FloatTensorGenerator(BaseWrapperTensorGenerator):
    r"""Implements a sequence generator that converts tensor values to
    float.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.tensor import Float, RandInt
        >>> generator = Float(RandInt(low=0, high=10))
        >>> generator
        FloatTensorGenerator(
          (tensor): RandIntTensorGenerator(low=0, high=10)
        )
        >>> generator.generate((2, 6))
        tensor([[...]])
    """

    def generate(self, size: tuple[int, ...], rng: Generator | None = None) -> Tensor:
        return self._generator.generate(size=size, rng=rng).float()


class LongTensorGenerator(BaseWrapperTensorGenerator):
    r"""Implements a sequence generator that converts a tensor values to
    long.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.tensor import Long, RandUniform
        >>> generator = Long(RandUniform(low=0, high=10))
        >>> generator
        LongTensorGenerator(
          (tensor): RandUniformTensorGenerator(low=0.0, high=10.0)
        )
        >>> generator.generate((2, 6))
        tensor([[...]])
    """

    def generate(self, size: tuple[int, ...], rng: Generator | None = None) -> Tensor:
        return self._generator.generate(size=size, rng=rng).long()



from __future__ import annotations

__all__ = ["MultinomialTensorGenerator", "UniformCategoricalTensorGenerator"]

import math
from collections.abc import Sequence

import torch
from torch import Generator, Tensor

from startorch.tensor.base import BaseTensorGenerator
from startorch.utils.weight import prepare_probabilities


class MultinomialTensorGenerator(BaseTensorGenerator):
    r"""Implements a class to generate tensors of categorical variables
    where each value is sampled from a multinomial distribution.

    Args:
    ----
        weights (``torch.Tensor`` of shape ``(num_categories,)`` and
            type float): Specifies the vector of weights associated
            at each category. The weights have to be positive but do
            not need to sum to 1.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from startorch.tensor import Multinomial
        >>> generator = Multinomial(torch.ones(10))
        >>> generator
        MultinomialTensorGenerator(num_categories=10)
        >>> generator.generate(size=(4, 12))
        tensor([[...]])
    """

    def __init__(self, weights: torch.Tensor | Sequence[int | float]) -> None:
        super().__init__()
        self._probabilities = prepare_probabilities(weights)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(num_categories={self._probabilities.numel()})"

    def generate(self, size: tuple[int, ...], rng: Generator | None = None) -> Tensor:
        return torch.multinomial(
            self._probabilities, math.prod(size), replacement=True, generator=rng
        ).view(*size)

    @classmethod
    def create_uniform_weights(cls, num_categories: int) -> MultinomialTensorGenerator:
        r"""Initializes the weights with a uniform pattern.

        All the categories have the same probability.
        The weight of the ``i``-th category (``w_i``) is generated
        with the rule: ``w_i = 1``

        Args:
        ----
            num_categories (int): Specifies the number of categories.

        Returns:
        -------
            ``MultinomialTensorGenerator``: A tensor generator where
                the weights of the multinomial distribution follow
                a uniform pattern.

        Example usage:

        .. code-block:: pycon

            >>> import torch
            >>> from startorch.tensor import Multinomial
            >>> generator = Multinomial.create_uniform_weights(10)
            >>> generator
            MultinomialTensorGenerator(num_categories=10)
            >>> generator.generate(size=(4, 12))
            tensor([[...]])
        """
        return cls(weights=torch.ones(num_categories))

    @classmethod
    def create_linear_weights(
        cls,
        num_categories: int,
    ) -> MultinomialTensorGenerator:
        r"""Initializes the weights with a linear pattern.

        The weight of the ``i``-th category (``w_i``) is generated
        with the rule: ``w_i = num_categories - i``

        Args:
        ----
            num_categories (int): Specifies the number of categories.

        Returns:
        -------
            ``MultinomialTensorGenerator``: A tensor generator where
                the weights of the multinomial distribution follow a
                linear pattern.

        Example usage:

        .. code-block:: pycon

            >>> import torch
            >>> from startorch.tensor import Multinomial
            >>> generator = Multinomial.create_linear_weights(10)
            >>> generator
            MultinomialTensorGenerator(num_categories=10)
            >>> generator.generate(size=(4, 12))
            tensor([[...]])
        """
        return cls(
            weights=num_categories * torch.ones(num_categories) - torch.arange(num_categories)
        )

    @classmethod
    def create_exp_weights(
        cls, num_categories: int, scale: float = 0.1
    ) -> MultinomialTensorGenerator:
        r"""Initializes the weights with an exponential pattern.

        The weight of the ``i``-th category (``w_i``) is generated with
        the rule: ``w_i = exp(-scale * i)``

        Args:
        ----
            num_categories (int): Specifies the number of categories.
            scale (float, optional): Specifies the scale parameter
                that controls the exponential function.
                Default: ``0.1``

        Returns:
        -------
            ``MultinomialTensorGenerator``: A tensor generator where
                the weights of the multinomial distribution follow
                an exponential pattern.

        Example usage:

        .. code-block:: pycon

            >>> import torch
            >>> from startorch.tensor import Multinomial
            >>> generator = Multinomial.create_exp_weights(10)
            >>> generator
            MultinomialTensorGenerator(num_categories=10)
            >>> generator.generate(size=(4, 12))
            tensor([[...]])
        """
        return cls(weights=torch.arange(num_categories).float().mul(-scale).exp())


class UniformCategoricalTensorGenerator(BaseTensorGenerator):
    r"""Implements a class to generate tensors of uniformly distributed
    categorical variables.

    All the categories have the same probability.

    Note: it is a more efficient implementation of
    ``Multinomial.generate_uniform_weights``.

    Args:
    ----
        num_categories (int): Specifies the number of categories.

    Raises:
    ------
        ValueError if ``num_categories`` is negative.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from startorch.tensor import UniformCategorical
        >>> generator = UniformCategorical(10)
        >>> generator
        UniformCategoricalTensorGenerator(num_categories=10)
        >>> generator.generate(size=(4, 12))
        tensor([[...]])
    """

    def __init__(self, num_categories: int) -> None:
        super().__init__()
        if num_categories <= 0:
            raise ValueError(
                f"num_categories has to be greater than 0 (received: {num_categories})"
            )
        self._num_categories = int(num_categories)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(num_categories={self._num_categories})"

    def generate(self, size: tuple[int, ...], rng: Generator | None = None) -> Tensor:
        return torch.randint(low=0, high=self._num_categories, size=size, generator=rng)



from __future__ import annotations

__all__ = ["RepeatPeriodicTimeSeriesGenerator"]

import math

from coola.utils import str_indent, str_mapping
from redcat import BatchDict
from torch import Generator

from startorch.periodic.timeseries import BasePeriodicTimeSeriesGenerator
from startorch.timeseries import BaseTimeSeriesGenerator, setup_timeseries_generator


class RepeatPeriodicTimeSeriesGenerator(BasePeriodicTimeSeriesGenerator):
    r"""Implements a class to generate periodic sequences by using a
    ``BaseTimeSeriesGenerator`` object and repeating the generated
    sequence.

    Args:
    ----
        sequence (``BaseTimeSeriesGenerator`` or dict): Specifies
            a sequence generator or its configuration.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.periodic.timeseries import Repeat
        >>> from startorch.timeseries import TimeSeries
        >>> from startorch.sequence import RandUniform
        >>> generator = Repeat(TimeSeries({"value": RandUniform(), "time": RandUniform()}))
        >>> generator
        RepeatPeriodicTimeSeriesGenerator(
          (generator): TimeSeriesGenerator(
              (value): RandUniformSequenceGenerator(low=0.0, high=1.0, feature_size=(1,))
              (time): RandUniformSequenceGenerator(low=0.0, high=1.0, feature_size=(1,))
            )
        )
        >>> generator.generate(seq_len=12, period=4, batch_size=4)
        BatchDict(
          (value): tensor([[...]], batch_dim=0, seq_dim=1)
          (time): tensor([[...]], batch_dim=0, seq_dim=1)
        )
    """

    def __init__(self, generator: BaseTimeSeriesGenerator | dict) -> None:
        super().__init__()
        self._generator = setup_timeseries_generator(generator)

    def __repr__(self) -> str:
        args = str_indent(str_mapping({"generator": self._generator}))
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def generate(
        self, seq_len: int, period: int, batch_size: int = 1, rng: Generator | None = None
    ) -> BatchDict:
        return (
            self._generator.generate(seq_len=period, batch_size=batch_size, rng=rng)
            .repeat_along_seq(math.ceil(seq_len / period))
            .slice_along_seq(stop=seq_len)
        )



from __future__ import annotations

__all__ = ["merge_multiple_timeseries_by_time", "truncate_timeseries_by_length"]

import torch

from startorch import constants as ct


def merge_multiple_timeseries_by_time(
    timeseries: list[dict[str, torch.Tensor]],
    time_key: str = ct.TIME,
) -> dict[str, torch.Tensor]:
    r"""Merges multiple time series by using the time.

    The sequence are merged based on the time information i.e. the
    merged time series is sorted by time. The input is a list of
    time series. Each time series is represented by a dictionary of
    ``torch.Tensor``s. All the values in the dictionary have to be
    of type ``torch.Tensor``. Each time series has to have a time
    tensor, and this time tensor is used to sort the merged time
    series. This function only supports the time tensor where
    the time is encoded as a single scalar (float, int, or long).
    The shape of the time tensor should be
    ``(batch_size, sequence_length, 1)`` or
    ``(batch_size, sequence_length)``. For example, the time in
    second can be used to encode the time information. All the
    tensors in each dictionary should have a shape
    ``(batch_size, sequence_length, *)`` where `*` means any
    number of dimensions. Only the sequence dimension can change
    between the time series. This function works for a variable
    number of features.

    Args:
    ----
        timeseries (list): Specifies the list of time series to merge.
            See the description above to know the format of the time
            series.
        time_key (str, optional): Specifies the key that contains the
            time information used to merge the time series.
            Default: ``'time'``

    Returns:
    -------
        dict: The merged time series. The dictionary has the same
            structure that the input time series.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from startorch import constants as ct
        >>> from startorch.timeseries import merge_multiple_timeseries_by_time
        >>> merge_multiple_timeseries_by_time(
        ...     [
        ...         {
        ...             ct.TIME: torch.tensor([[[5], [10], [15], [20], [25]]], dtype=torch.float),
        ...             ct.VALUE: torch.tensor([[11, 12, 13, 14, 15]], dtype=torch.long),
        ...         },
        ...         {
        ...             ct.TIME: torch.tensor([[[6], [12], [16], [24]]], dtype=torch.float),
        ...             ct.VALUE: torch.tensor([[21, 22, 23, 24]], dtype=torch.long),
        ...         },
        ...     ]
        ... )
        {'time': tensor([[[ 5.],
                 [ 6.],
                 [10.],
                 [12.],
                 [15.],
                 [16.],
                 [20.],
                 [24.],
                 [25.]]]), 'value': tensor([[11, 21, 12, 22, 13, 23, 14, 24, 15]])}
    """
    all_time_second = torch.cat([ts[time_key] for ts in timeseries], dim=1)
    sorted_values, indices = torch.sort(all_time_second, dim=1, stable=True)
    output = {time_key: sorted_values}

    # Merge the other keys.
    keys = set(timeseries[0].keys())
    keys.remove(time_key)
    for key in keys:
        inputs = torch.cat([ts[key] for ts in timeseries], dim=1)
        shapes = [1] * inputs.ndim
        shapes[:2] = inputs.shape[:2]
        output[key] = torch.gather(
            inputs, dim=1, index=indices.view(*shapes).repeat(1, 1, *inputs.shape[2:])
        )

    return output


def truncate_timeseries_by_length(
    timeseries: dict[str, torch.Tensor], max_seq_len: int
) -> dict[str, torch.Tensor]:
    r"""Truncates a time series by a maximum length.

    Each time series is represented by a dictionary of
    ``torch.Tensor``s. All the values in the dictionary have to be of
    type ``torch.Tensor``. All the tensors in each dictionary should
    have a shape ``(batch_size, sequence_length, *)`` where `*` means
    any number of dimensions. This function works for a variable
    number of features (keys).

    Args:
    ----
        timeseries (dict): Specifies the time series to truncate.
        max_seq_len (int): Specifies the maximum length of the
            time series.

    Returns:
    -------
        dict: The truncated time series where the sequence length
            is at most ``max_seq_len``.

    Raises:
    ------
        RuntimeError if ``max_seq_len`` is lowe than 1.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from startorch import constants as ct
        >>> from startorch.timeseries import truncate_timeseries_by_length
        >>> truncate_timeseries_by_length(
        ...     {
        ...         ct.TIME: torch.tensor([[[5], [10], [15], [20], [25]]], dtype=torch.float),
        ...         ct.VALUE: torch.tensor([[11, 12, 13, 14, 15]], dtype=torch.long),
        ...     },
        ...     max_seq_len=3,
        ... )
        {'time': tensor([[[ 5.],
                 [10.],
                 [15.]]]), 'value': tensor([[11, 12, 13]])}
    """
    if max_seq_len < 1:
        raise RuntimeError(
            f"`max_seq_len` has to be greater or equal to 1 (received: {max_seq_len:,})"
        )
    truncated_timeseries = {}
    for key, value in timeseries.items():
        truncated_timeseries[key] = value[:, :max_seq_len]
    return truncated_timeseries



from __future__ import annotations

__all__ = ["RandWienerProcessSequenceGenerator", "wiener_process"]

import math

import torch
from redcat import BatchedTensorSeq
from torch import Generator, Tensor

from startorch.sequence.base import BaseSequenceGenerator


class RandWienerProcessSequenceGenerator(BaseSequenceGenerator):
    r"""Implements a sequence generator to generate sequences where the
    values are generated by a Wiener process.

    Useful link: https://en.wikipedia.org/wiki/Wiener_process

    Args:
    ----
        step_size (float, optional): Specifies the time
            step size. Default: ``1.0``

    Raises:
    ------
        ValueError if ``step_size`` is not a positive number.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.sequence import RandWienerProcess
        >>> generator = RandWienerProcess()
        >>> generator
        RandWienerProcessSequenceGenerator(time_step_size=1.0)
        >>> generator.generate(seq_len=6, batch_size=2)
        tensor([[...]], batch_dim=0, seq_dim=1)
    """

    def __init__(self, step_size: float = 1.0) -> None:
        super().__init__()
        if step_size < 0:
            raise ValueError(f"step_size has to be greater than 0 (received: {step_size})")
        self._step_size = float(step_size)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(time_step_size={self._step_size})"

    def generate(
        self, seq_len: int, batch_size: int = 1, rng: Generator | None = None
    ) -> BatchedTensorSeq:
        return BatchedTensorSeq(
            wiener_process(
                seq_len=seq_len,
                batch_size=batch_size,
                step_size=self._step_size,
                generator=rng,
            ).unsqueeze(dim=2)
        )


def wiener_process(
    seq_len: int,
    step_size: float = 1.0,
    batch_size: int = 1,
    generator: Generator | None = None,
) -> Tensor:
    r"""Creates a batch of sequences generated by a Wiener process.

    Useful link: https://en.wikipedia.org/wiki/Wiener_process

    Args:
    ----
        seq_len (int): Specifies the sequence length.
        step_size (float, optional): Specifies the time
            step size. Default: ``1.0``
        batch_size (int, optional): Specifies the batch size.
            Default: ``1``
        generator (``torch.Generator`` or None): Specifies an optional
            random generator. Default: ``None``

    Returns:
    -------
        ``torch.Tensor`` of shape ``(batch size, sequence length)``
            and type float: A batch of Wiener processes.

    Raises:
    ------
        ValueError if ``step_size`` is not a positive number.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.sequence.wiener import wiener_process
        >>> wiener_process(seq_len=12, batch_size=4)
        tensor([[...]])
    """
    if step_size < 0:
        raise ValueError(f"step_size has to be greater than 0 (received: {step_size})")
    increments = torch.randn((batch_size, seq_len), generator=generator).mul(math.sqrt(step_size))
    increments[:, 0] = 0
    return torch.cumsum(increments, dim=1)



from __future__ import annotations

__all__ = ["matplotlib_available"]

from pytest import mark

from startorch.utils.imports import is_matplotlib_available

matplotlib_available = mark.skipif(not is_matplotlib_available(), reason="Requires matplotlib")



from __future__ import annotations

__all__ = ["FloatSequenceGenerator", "LongSequenceGenerator"]

from redcat import BatchedTensorSeq
from torch import Generator

from startorch.sequence.wrapper import BaseWrapperSequenceGenerator


class FloatSequenceGenerator(BaseWrapperSequenceGenerator):
    r"""Implements a sequence generator that converts a batch of
    sequences to float type.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.sequence import Float, RandInt
        >>> generator = Float(RandInt(low=0, high=10))
        >>> generator
        FloatSequenceGenerator(
          (sequence): RandIntSequenceGenerator(low=0, high=10, feature_size=())
        )
        >>> generator.generate(seq_len=6, batch_size=2)
        tensor([[...]], batch_dim=0, seq_dim=1)
    """

    def generate(
        self, seq_len: int, batch_size: int = 1, rng: Generator | None = None
    ) -> BatchedTensorSeq:
        return self._generator.generate(seq_len=seq_len, batch_size=batch_size, rng=rng).float()


class LongSequenceGenerator(BaseWrapperSequenceGenerator):
    r"""Implements a sequence generator that converts a batch of
    sequences to long type.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.sequence import Long, RandUniform
        >>> generator = Long(RandUniform(low=0, high=10))
        >>> generator
        LongSequenceGenerator(
          (sequence): RandUniformSequenceGenerator(low=0.0, high=10.0, feature_size=(1,))
        )
        >>> generator.generate(seq_len=6, batch_size=2)
        tensor([[...]], batch_dim=0, seq_dim=1)
    """

    def generate(
        self, seq_len: int, batch_size: int = 1, rng: Generator | None = None
    ) -> BatchedTensorSeq:
        return self._generator.generate(seq_len=seq_len, batch_size=batch_size, rng=rng).long()



from __future__ import annotations

__all__ = [
    "BasePeriodicSequenceGenerator",
    "is_periodic_sequence_generator_config",
    "setup_periodic_sequence_generator",
]

import logging
from abc import ABC, abstractmethod

from objectory import AbstractFactory
from objectory.utils import is_object_config
from redcat import BatchedTensorSeq
from torch import Generator

from startorch.utils.format import str_target_object

logger = logging.getLogger(__name__)


class BasePeriodicSequenceGenerator(ABC, metaclass=AbstractFactory):
    r"""Defines the base class to generate periodic sequences.

    A child class has to implement the ``generate`` method.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from startorch.periodic.sequence import Repeat
        >>> from startorch.sequence import RandUniform
        >>> generator = Repeat(RandUniform())
        >>> generator
        RepeatPeriodicSequenceGenerator(
          (generator): RandUniformSequenceGenerator(low=0.0, high=1.0, feature_size=(1,))
        )
        >>> generator.generate(seq_len=12, period=4, batch_size=4)
        tensor([[...]], batch_dim=0, seq_dim=1)
    """

    @abstractmethod
    def generate(
        self, seq_len: int, period: int, batch_size: int = 1, rng: Generator | None = None
    ) -> BatchedTensorSeq:
        r"""Generates a batch of periodic sequences.

        All the sequences in the batch have the same length.

        Args:
        ----
            seq_len (int): Specifies the sequence length.
            period (int): Specifies the period.
            batch_size (int, optional): Specifies the batch size.
                Default: ``1``
            rng (``torch.Generator`` or None, optional): Specifies
                an optional random number generator. Default: ``None``

        Returns:
        -------
            ``BatchedTensorSeq``: A batch of sequences. The data in the
                batch are represented by a ``torch.Tensor`` of shape
                ``(batch_size, sequence_length, *)`` where `*` means
                any number of dimensions.

        Example usage:

        .. code-block:: pycon

            >>> import torch
            >>> from startorch.periodic.sequence import Repeat
            >>> from startorch.sequence import RandUniform
            >>> generator = Repeat(RandUniform())
            >>> generator.generate(seq_len=12, period=4, batch_size=4)
            tensor([[...]], batch_dim=0, seq_dim=1)
        """


def is_periodic_sequence_generator_config(config: dict) -> bool:
    r"""Indicates if the input configuration is a configuration for a
    ``BasePeriodicSequenceGenerator``.

    This function only checks if the value of the key  ``_target_``
    is valid. It does not check the other values. If ``_target_``
    indicates a function, the returned type hint is used to check
    the class.

    Args:
    ----
        config (dict): Specifies the configuration to check.

    Returns:
    -------
        bool: ``True`` if the input configuration is a configuration
            for a ``BasePeriodicSequenceGenerator`` object.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.periodic.sequence import is_periodic_sequence_generator_config
        >>> is_periodic_sequence_generator_config(
        ...     {
        ...         "_target_": "startorch.periodic.sequence.Repeat",
        ...         "generator": {"_target_": "startorch.sequence.RandUniform"},
        ...     }
        ... )
        True
    """
    return is_object_config(config, BasePeriodicSequenceGenerator)


def setup_periodic_sequence_generator(
    generator: BasePeriodicSequenceGenerator | dict,
) -> BasePeriodicSequenceGenerator:
    r"""Sets up a periodic sequence generator.

    The sequence generator is instantiated from its configuration by
    using the ``BasePeriodicSequenceGenerator`` factory function.

    Args:
    ----
        generator (``BasePeriodicSequenceGenerator`` or dict): Specifies a
            periodic sequence generator or its configuration.

    Returns:
    -------
        ``BasePeriodicSequenceGenerator``: A periodic sequence generator.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.periodic.sequence import setup_periodic_sequence_generator
        >>> setup_periodic_sequence_generator(
        ...     {
        ...         "_target_": "startorch.periodic.sequence.Repeat",
        ...         "generator": {"_target_": "startorch.sequence.RandUniform"},
        ...     }
        ... )
        RepeatPeriodicSequenceGenerator(
          (generator): RandUniformSequenceGenerator(low=0.0, high=1.0, feature_size=(1,))
        )
    """
    if isinstance(generator, dict):
        logger.info(
            "Initializing a periodic sequence generator from its configuration... "
            f"{str_target_object(generator)}"
        )
        generator = BasePeriodicSequenceGenerator.factory(**generator)
    if not isinstance(generator, BasePeriodicSequenceGenerator):
        logger.warning(
            f"generator is not a `BasePeriodicSequenceGenerator` (received: {type(generator)})"
        )
    return generator



from __future__ import annotations

__all__ = [
    "Abs",
    "Acosh",
    "Add",
    "AddScalar",
    "Asinh",
    "AsinhUniform",
    "Atanh",
    "BaseTensorGenerator",
    "BaseWrapperTensorGenerator",
    "Cauchy",
    "Clamp",
    "Cosh",
    "Div",
    "Exp",
    "Exponential",
    "Float",
    "Fmod",
    "Full",
    "HalfCauchy",
    "HalfNormal",
    "Log",
    "LogNormal",
    "LogUniform",
    "Long",
    "Mul",
    "MulScalar",
    "Multinomial",
    "MultinomialChoice",
    "Neg",
    "Normal",
    "Poisson",
    "RandAsinhUniform",
    "RandCauchy",
    "RandExponential",
    "RandHalfCauchy",
    "RandHalfNormal",
    "RandInt",
    "RandLogNormal",
    "RandLogUniform",
    "RandNormal",
    "RandPoisson",
    "RandTruncCauchy",
    "RandTruncExponential",
    "RandTruncHalfCauchy",
    "RandTruncHalfNormal",
    "RandTruncLogNormal",
    "RandTruncNormal",
    "RandUniform",
    "Sinh",
    "Sqrt",
    "Sub",
    "Tanh",
    "TruncCauchy",
    "TruncExponential",
    "TruncHalfCauchy",
    "TruncHalfNormal",
    "TruncLogNormal",
    "TruncNormal",
    "Uniform",
    "UniformCategorical",
    "is_tensor_generator_config",
    "setup_tensor_generator",
]

from startorch.tensor.base import (
    BaseTensorGenerator,
    is_tensor_generator_config,
    setup_tensor_generator,
)
from startorch.tensor.categorical import MultinomialTensorGenerator as Multinomial
from startorch.tensor.categorical import (
    UniformCategoricalTensorGenerator as UniformCategorical,
)
from startorch.tensor.cauchy import CauchyTensorGenerator as Cauchy
from startorch.tensor.cauchy import RandCauchyTensorGenerator as RandCauchy
from startorch.tensor.cauchy import RandTruncCauchyTensorGenerator as RandTruncCauchy
from startorch.tensor.cauchy import TruncCauchyTensorGenerator as TruncCauchy
from startorch.tensor.choice import (
    MultinomialChoiceTensorGenerator as MultinomialChoice,
)
from startorch.tensor.constant import FullTensorGenerator as Full
from startorch.tensor.dtype import FloatTensorGenerator as Float
from startorch.tensor.dtype import LongTensorGenerator as Long
from startorch.tensor.exponential import ExponentialTensorGenerator as Exponential
from startorch.tensor.exponential import (
    RandExponentialTensorGenerator as RandExponential,
)
from startorch.tensor.exponential import (
    RandTruncExponentialTensorGenerator as RandTruncExponential,
)
from startorch.tensor.exponential import (
    TruncExponentialTensorGenerator as TruncExponential,
)
from startorch.tensor.halfcauchy import HalfCauchyTensorGenerator as HalfCauchy
from startorch.tensor.halfcauchy import RandHalfCauchyTensorGenerator as RandHalfCauchy
from startorch.tensor.halfcauchy import (
    RandTruncHalfCauchyTensorGenerator as RandTruncHalfCauchy,
)
from startorch.tensor.halfcauchy import (
    TruncHalfCauchyTensorGenerator as TruncHalfCauchy,
)
from startorch.tensor.halfnormal import HalfNormalTensorGenerator as HalfNormal
from startorch.tensor.halfnormal import RandHalfNormalTensorGenerator as RandHalfNormal
from startorch.tensor.halfnormal import (
    RandTruncHalfNormalTensorGenerator as RandTruncHalfNormal,
)
from startorch.tensor.halfnormal import (
    TruncHalfNormalTensorGenerator as TruncHalfNormal,
)
from startorch.tensor.lognormal import LogNormalTensorGenerator as LogNormal
from startorch.tensor.lognormal import RandLogNormalTensorGenerator as RandLogNormal
from startorch.tensor.lognormal import (
    RandTruncLogNormalTensorGenerator as RandTruncLogNormal,
)
from startorch.tensor.lognormal import TruncLogNormalTensorGenerator as TruncLogNormal
from startorch.tensor.math import AbsTensorGenerator as Abs
from startorch.tensor.math import AddScalarTensorGenerator as AddScalar
from startorch.tensor.math import AddTensorGenerator as Add
from startorch.tensor.math import ClampTensorGenerator as Clamp
from startorch.tensor.math import DivTensorGenerator as Div
from startorch.tensor.math import ExpTensorGenerator as Exp
from startorch.tensor.math import FmodTensorGenerator as Fmod
from startorch.tensor.math import LogTensorGenerator as Log
from startorch.tensor.math import MulScalarTensorGenerator as MulScalar
from startorch.tensor.math import MulTensorGenerator as Mul
from startorch.tensor.math import NegTensorGenerator as Neg
from startorch.tensor.math import SqrtTensorGenerator as Sqrt
from startorch.tensor.math import SubTensorGenerator as Sub
from startorch.tensor.normal import NormalTensorGenerator as Normal
from startorch.tensor.normal import RandNormalTensorGenerator as RandNormal
from startorch.tensor.normal import RandTruncNormalTensorGenerator as RandTruncNormal
from startorch.tensor.normal import TruncNormalTensorGenerator as TruncNormal
from startorch.tensor.poisson import PoissonTensorGenerator as Poisson
from startorch.tensor.poisson import RandPoissonTensorGenerator as RandPoisson
from startorch.tensor.trigo import AcoshTensorGenerator as Acosh
from startorch.tensor.trigo import AsinhTensorGenerator as Asinh
from startorch.tensor.trigo import AtanhTensorGenerator as Atanh
from startorch.tensor.trigo import CoshTensorGenerator as Cosh
from startorch.tensor.trigo import SinhTensorGenerator as Sinh
from startorch.tensor.trigo import TanhTensorGenerator as Tanh
from startorch.tensor.uniform import AsinhUniformTensorGenerator as AsinhUniform
from startorch.tensor.uniform import LogUniformTensorGenerator as LogUniform
from startorch.tensor.uniform import RandAsinhUniformTensorGenerator as RandAsinhUniform
from startorch.tensor.uniform import RandIntTensorGenerator as RandInt
from startorch.tensor.uniform import RandLogUniformTensorGenerator as RandLogUniform
from startorch.tensor.uniform import RandUniformTensorGenerator as RandUniform
from startorch.tensor.uniform import UniformTensorGenerator as Uniform
from startorch.tensor.wrapper import BaseWrapperTensorGenerator



from __future__ import annotations

__all__ = ["MultinomialChoiceTensorGenerator"]

from collections.abc import Sequence

import torch
from coola.utils.format import str_indent
from torch import Generator, Tensor

from startorch.tensor.base import BaseTensorGenerator, setup_tensor_generator
from startorch.utils.format import str_weighted_modules
from startorch.utils.weight import prepare_weighted_generators


class MultinomialChoiceTensorGenerator(BaseTensorGenerator):
    r"""Implements a tensor generator that select a tensor generator at
    each batch.

    This tensor generator is used to generate tensors with different
    generation processes. The user can specify a list of tensor
    generators with an associated weight. The weight is used to sample
    the tensor generator with a multinomial distribution. Higher
    weight means that the tensor generator has a higher probability
    to be selected at each batch. Each dictionary in the
    ``generators`` input should have the following items:

        - a key ``'generator'`` which indicates the tensor generator
            or its configuration.
        - an optional key ``'weight'`` with a float value which
            indicates the weight of the tensor generator.
            If this key is absent, the weight is set to ``1.0``.

    Args:
    ----
        generators (tuple or list): Specifies the tensor
            generators and their weights. See above to learn
            about the expected format.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.tensor import MultinomialChoice, RandUniform, RandNormal
        >>> generator = MultinomialChoice(
        ...     (
        ...         {"weight": 2.0, "generator": RandUniform()},
        ...         {"weight": 1.0, "generator": RandNormal()},
        ...     )
        ... )
        >>> generator
        MultinomialChoiceTensorGenerator(
          (0) [weight=2.0] RandUniformTensorGenerator(low=0.0, high=1.0)
          (1) [weight=1.0] RandNormalTensorGenerator(mean=0.0, std=1.0)
        )
        >>> generator.generate(size=(4, 12))
        tensor([[...]])
    """

    def __init__(self, generators: Sequence[dict[str, BaseTensorGenerator | dict]]) -> None:
        super().__init__()
        generators, weights = prepare_weighted_generators(generators)
        self._generators = tuple(setup_tensor_generator(generator) for generator in generators)
        self._weights = torch.as_tensor(weights, dtype=torch.float)

    def __repr__(self) -> str:
        args = str_indent(str_weighted_modules(modules=self._generators, weights=self._weights))
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def generate(self, size: tuple[int, ...], rng: Generator | None = None) -> Tensor:
        index = torch.multinomial(self._weights, num_samples=1, generator=rng).item()
        return self._generators[index].generate(size=size, rng=rng)



from __future__ import annotations

__all__ = ["SineWaveSequenceGenerator"]

import math

from coola.utils.format import str_indent, str_mapping
from redcat import BatchedTensorSeq
from torch import Generator

from startorch.sequence.base import BaseSequenceGenerator, setup_sequence_generator


class SineWaveSequenceGenerator(BaseSequenceGenerator):
    r"""Implements a class to generate sequence by sampling values with a
    sine wave pattern.

    The sequences are generated by using the following formula:

    ``output = amplitude * sin(2 * pi * frequency * value + phase)``

    Args:
        value (``BaseSequenceGenerator`` or dict): Specifies a
            sequence generator (or its configuration) to generate the
            sequence values.
        frequency (``BaseSequenceGenerator`` or dict): Specifies a
            sequence generator (or its configuration) to generate the
            frequency values.
        phase (``BaseSequenceGenerator`` or dict): Specifies a
            sequence generator (or its configuration) to generate the
            phase values.
        amplitude (``BaseSequenceGenerator`` or dict): Specifies a
            sequence generator (or its configuration) to generate the
            amplitude values.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from startorch.sequence import Arange, SineWave, RandUniform, Constant, RandLogUniform
        >>> generator = SineWave(
        ...     value=Arange(),
        ...     frequency=Constant(RandLogUniform(low=0.01, high=0.1)),
        ...     phase=Constant(RandUniform(low=-1.0, high=1.0)),
        ...     amplitude=Constant(RandLogUniform(low=0.1, high=1.0)),
        ... )
        >>> generator
        SineWaveSequenceGenerator(
          (value): ArangeSequenceGenerator(feature_size=(1,))
          (frequency): ConstantSequenceGenerator(
              (sequence): RandLogUniformSequenceGenerator(low=0.01, high=0.1, feature_size=(1,))
            )
          (phase): ConstantSequenceGenerator(
              (sequence): RandUniformSequenceGenerator(low=-1.0, high=1.0, feature_size=(1,))
            )
          (amplitude): ConstantSequenceGenerator(
              (sequence): RandLogUniformSequenceGenerator(low=0.1, high=1.0, feature_size=(1,))
            )
        )
        >>> generator.generate(seq_len=12, batch_size=4)
        tensor([[...]], batch_dim=0, seq_dim=1)
    """

    def __init__(
        self,
        value: BaseSequenceGenerator | dict,
        frequency: BaseSequenceGenerator | dict,
        phase: BaseSequenceGenerator | dict,
        amplitude: BaseSequenceGenerator | dict,
    ) -> None:
        super().__init__()
        self._value = setup_sequence_generator(value)
        self._frequency = setup_sequence_generator(frequency)
        self._phase = setup_sequence_generator(phase)
        self._amplitude = setup_sequence_generator(amplitude)

    def __repr__(self) -> str:
        args = str_indent(
            str_mapping(
                {
                    "value": self._value,
                    "frequency": self._frequency,
                    "phase": self._phase,
                    "amplitude": self._amplitude,
                }
            )
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def generate(
        self, seq_len: int, batch_size: int = 1, rng: Generator | None = None
    ) -> BatchedTensorSeq:
        return (
            self._value.generate(seq_len=seq_len, batch_size=batch_size, rng=rng)
            .mul(self._frequency.generate(seq_len=seq_len, batch_size=batch_size, rng=rng))
            .mul(2.0 * math.pi)
            .add(self._phase.generate(seq_len=seq_len, batch_size=batch_size, rng=rng))
            .sin()
            .mul(self._amplitude.generate(seq_len=seq_len, batch_size=batch_size, rng=rng))
        )



from __future__ import annotations

__all__ = ["BaseWrapperTensorGenerator"]


from coola.utils.format import str_indent, str_mapping

from startorch.tensor.base import BaseTensorGenerator, setup_tensor_generator


class BaseWrapperTensorGenerator(BaseTensorGenerator):
    r"""Defines a base class to easily wrap a tensor generator.

    Args:
    ----
        tensor (``BaseTensorGenerator`` or dict):
            Specifies the tensor generator or its configuration.
    """

    def __init__(self, generator: BaseTensorGenerator | dict) -> None:
        super().__init__()
        self._generator = setup_tensor_generator(generator)

    def __repr__(self) -> str:
        args = str_indent(str_mapping({"tensor": self._generator}))
        return f"{self.__class__.__qualname__}(\n  {args}\n)"



from __future__ import annotations

__all__ = ["AutoRegressiveSequenceGenerator"]


import torch
from coola.utils.format import str_indent, str_mapping
from redcat import BatchedTensorSeq
from torch import Generator

from startorch.sequence.base import BaseSequenceGenerator, setup_sequence_generator
from startorch.tensor.base import BaseTensorGenerator, setup_tensor_generator


class AutoRegressiveSequenceGenerator(BaseSequenceGenerator):
    r"""Implements a class to generate sequence by sampling values with a
    linear pattern.

    Args:
    ----
        value (``BaseSequenceGenerator`` or dict): Specifies a
            sequence generator (or its configuration) used to generate
            the initial sequence values. These values are used to
            start the AR.
        coefficient (``BaseSequenceGenerator`` or dict): Specifies a
            sequence generator (or its configuration) used to generate
            the coefficients.
        noise (``BaseSequenceGenerator`` or dict): Specifies a sequence
            generator (or its configuration) used to generate the
            noise values.
        order (``BaseTensorGenerator`` or dict): Specifies a tensor
            generator (or its configuration) used to generate the order
            of the AR.
        max_abs_value (float): Specifies the maximum absolute value.
            This argument ensures the values stay in the range
            ``[-max_abs_value, max_abs_value]``.
        warmup (int, optional): Specifies the number of cycles used to
            initiate the AR. The initial value sampled do not follow
            an AR, so using warmup allows to initialize the AR so each
            value follows an AR. Default: ``10``

    Raises:
    ------
        ValueError if ``max_abs_value`` is not a positive number.
        ValueError if ``warmup`` is not a positive number.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from startorch.sequence import AutoRegressive, RandUniform, RandNormal, Full
        >>> from startorch.tensor import RandInt
        >>> generator = AutoRegressive(
        ...     value=RandNormal(),
        ...     coefficient=RandUniform(low=-1.0, high=1.0),
        ...     noise=Full(0.0),
        ...     order=RandInt(low=1, high=6),
        ...     max_abs_value=100.0,
        ... )
        >>> generator
        AutoRegressiveSequenceGenerator(
          (value): RandNormalSequenceGenerator(mean=0.0, std=1.0, feature_size=(1,))
          (coefficient): RandUniformSequenceGenerator(low=-1.0, high=1.0, feature_size=(1,))
          (noise): FullSequenceGenerator(value=0.0, feature_size=(1,))
          (order): RandIntTensorGenerator(low=1, high=6)
          (max_abs_value): 100.0
          (warmup): 10
        )
        >>> generator.generate(seq_len=12, batch_size=4)
        tensor([[...]], batch_dim=0, seq_dim=1)
    """

    def __init__(
        self,
        value: BaseSequenceGenerator | dict,
        coefficient: BaseSequenceGenerator | dict,
        noise: BaseSequenceGenerator | dict,
        order: BaseTensorGenerator | dict,
        max_abs_value: float,
        warmup: int = 10,
    ) -> None:
        super().__init__()
        self._value = setup_sequence_generator(value)
        self._coefficient = setup_sequence_generator(coefficient)
        self._noise = setup_sequence_generator(noise)
        self._order = setup_tensor_generator(order)

        if max_abs_value <= 0.0:
            raise ValueError(f"`max_abs_value` has to be positive but received {max_abs_value}")
        self._max_abs_value = float(max_abs_value)
        if warmup < 0:
            raise ValueError(f"warmup has to be positive or zero but received {warmup}")
        self._warmup = int(warmup)

    def __repr__(self) -> str:
        args = str_indent(
            str_mapping(
                {
                    "value": self._value,
                    "coefficient": self._coefficient,
                    "noise": self._noise,
                    "order": self._order,
                    "max_abs_value": self._max_abs_value,
                    "warmup": self._warmup,
                }
            )
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def generate(
        self, seq_len: int, batch_size: int = 1, rng: Generator | None = None
    ) -> BatchedTensorSeq:
        order = int(self._order.generate((1,), rng=rng).item())
        if order < 1:
            raise RuntimeError(f"Order must be a positive integer but received {order}")
        x = self._value.generate(
            seq_len=seq_len + order * self._warmup, batch_size=batch_size, rng=rng
        ).data
        noise = self._noise.generate(
            seq_len=seq_len + order * self._warmup, batch_size=batch_size, rng=rng
        ).data
        coeffs = self._coefficient.generate(seq_len=order, batch_size=batch_size, rng=rng).data
        for i in range(order, seq_len + order * self._warmup):
            x[:, i] = torch.fmod(
                torch.sum(coeffs * x[:, i - order : i], dim=1) + noise[:, i], self._max_abs_value
            )
        return BatchedTensorSeq(x).slice_along_seq(order * self._warmup)



from __future__ import annotations

__all__ = ["MultinomialChoiceSequenceGenerator"]

from collections.abc import Sequence

import torch
from coola.utils.format import str_indent
from redcat import BatchedTensorSeq
from torch import Generator

from startorch.sequence.base import BaseSequenceGenerator, setup_sequence_generator
from startorch.utils.format import str_weighted_modules
from startorch.utils.weight import prepare_weighted_generators


class MultinomialChoiceSequenceGenerator(BaseSequenceGenerator):
    r"""Implements a sequence generator that select a sequence generator
    at each batch.

    This sequence generator is used to generate sequences with different
    generation processes. The user can specify a list of sequence
    generators with an associated weight. The weight is used to sample
    the sequence generator with a multinomial distribution. Higher
    weight means that the sequence generator has a higher probability
    to be selected at each batch. Each dictionary in the
    ``generators`` input should have the following items:

        - a key ``'generator'`` which indicates the sequence generator
            or its configuration.
        - an optional key ``'weight'`` with a float value which
            indicates the weight of the sequence generator.
            If this key is absent, the weight is set to ``1.0``.

    Args:
    ----
        sequences (sequence): Specifies the sequence
            generators and their weights. See above to learn
            about the expected format.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.sequence import MultinomialChoice, RandUniform, RandNormal
        >>> generator = MultinomialChoice(
        ...     (
        ...         {"weight": 2.0, "generator": RandUniform()},
        ...         {"weight": 1.0, "generator": RandNormal()},
        ...     )
        ... )
        >>> generator.generate(seq_len=10, batch_size=2)
        tensor([[...]], batch_dim=0, seq_dim=1)
    """

    def __init__(self, sequences: Sequence[dict[str, BaseSequenceGenerator | dict]]) -> None:
        super().__init__()
        sequences, weights = prepare_weighted_generators(sequences)
        self._sequences = tuple(setup_sequence_generator(generator) for generator in sequences)
        self._weights = torch.as_tensor(weights, dtype=torch.float)

    def __repr__(self) -> str:
        args = str_indent(str_weighted_modules(modules=self._sequences, weights=self._weights))
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def generate(
        self, seq_len: int, batch_size: int = 1, rng: Generator | None = None
    ) -> BatchedTensorSeq:
        index = torch.multinomial(self._weights, num_samples=1, generator=rng).item()
        return self._sequences[index].generate(seq_len=seq_len, batch_size=batch_size, rng=rng)



from __future__ import annotations

__all__ = [
    "HalfNormalTensorGenerator",
    "RandHalfNormalTensorGenerator",
    "RandTruncHalfNormalTensorGenerator",
    "TruncHalfNormalTensorGenerator",
]


from coola.utils.format import str_indent, str_mapping
from torch import Generator, Tensor

from startorch.random import (
    half_normal,
    rand_half_normal,
    rand_trunc_half_normal,
    trunc_half_normal,
)
from startorch.tensor.base import BaseTensorGenerator, setup_tensor_generator


class HalfNormalTensorGenerator(BaseTensorGenerator):
    r"""Implements a class to generate tensor by sampling values from a
    half-Normal distribution.

    Args:
    ----
        std (``BaseTensorGenerator`` or dict): Specifies a tensor
            generator (or its configuration) to generate the standard
            deviation.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.tensor import HalfNormal, RandUniform
        >>> generator = HalfNormal(std=RandUniform(low=1.0, high=2.0))
        >>> generator
        HalfNormalTensorGenerator(
          (std): RandUniformTensorGenerator(low=1.0, high=2.0)
        )
        >>> generator.generate(size=(2, 6))
        tensor([[...]])
    """

    def __init__(self, std: BaseTensorGenerator | dict) -> None:
        super().__init__()
        self._std = setup_tensor_generator(std)

    def __repr__(self) -> str:
        args = str_indent(str_mapping({"std": self._std}))
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def generate(self, size: tuple[int, ...], rng: Generator | None = None) -> Tensor:
        return half_normal(
            std=self._std.generate(size=size, rng=rng),
            generator=rng,
        )


class RandHalfNormalTensorGenerator(BaseTensorGenerator):
    r"""Implements a class to generate tensor by sampling values from a
    half-Normal distribution.

    Args:
    ----
        std (float, optional): Specifies the std of the distribution.
            Default: ``1.0``

    Raises:
    ------
        ValueError if ``std`` is not a positive number.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.tensor import RandHalfNormal
        >>> generator = RandHalfNormal(std=1.0)
        >>> generator
        RandHalfNormalTensorGenerator(std=1.0)
        >>> generator.generate(size=(2, 6))
        tensor([[...]])
    """

    def __init__(self, std: float = 1.0) -> None:
        super().__init__()
        if std <= 0:
            raise ValueError(f"std has to be greater than 0 (received: {std})")
        self._std = float(std)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(std={self._std})"

    def generate(self, size: tuple[int, ...], rng: Generator | None = None) -> Tensor:
        return rand_half_normal(
            size=size,
            std=self._std,
            generator=rng,
        )


class RandTruncHalfNormalTensorGenerator(BaseTensorGenerator):
    r"""Implements a class to generate tensor by sampling values from a
    truncated half-Normal distribution.

    Args:
    ----
        std (float, optional): Specifies the std of the distribution.
            Default: ``1.0``
        max_value (float, optional): Specifies the maximum value.
            Default: ``3.0``

    Raises:
    ------
        ValueError if ``std`` is not a positive number.
        ValueError if ``max_value`` is not a positive number.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.tensor import RandTruncHalfNormal
        >>> generator = RandTruncHalfNormal(std=1.0, max_value=1.0)
        >>> generator
        RandTruncHalfNormalTensorGenerator(std=1.0, max_value=1.0)
        >>> generator.generate(size=(2, 6))
        tensor([[...]])
    """

    def __init__(self, std: float = 1.0, max_value: float = 3.0) -> None:
        super().__init__()
        if std <= 0:
            raise ValueError(f"std has to be greater than 0 (received: {std})")
        self._std = float(std)
        if max_value <= 0:
            raise ValueError(f"max_value has to be greater than 0 (received: {max_value})")
        self._max_value = float(max_value)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(std={self._std}, max_value={self._max_value})"

    def generate(self, size: tuple[int, ...], rng: Generator | None = None) -> Tensor:
        return rand_trunc_half_normal(
            size=size,
            std=self._std,
            max_value=self._max_value,
            generator=rng,
        )


class TruncHalfNormalTensorGenerator(BaseTensorGenerator):
    r"""Implements a class to generate tensor by sampling values from a
    half-Normal distribution.

    Args:
    ----
        std (``BaseTensorGenerator`` or dict): Specifies a tensor
            generator (or its configuration) to generate the std.
        max_value (``BaseTensorGenerator`` or dict): Specifies a
            tensor generator (or its configuration) to generate the
            maximum value (excluded).

    Example usage:

    .. code-block:: pycon

        >>> from startorch.tensor import RandUniform, TruncHalfNormal
        >>> generator = TruncHalfNormal(
        ...     std=RandUniform(low=1.0, high=2.0),
        ...     max_value=RandUniform(low=5.0, high=10.0),
        ... )
        >>> generator
        TruncHalfNormalTensorGenerator(
          (std): RandUniformTensorGenerator(low=1.0, high=2.0)
          (max_value): RandUniformTensorGenerator(low=5.0, high=10.0)
        )
        >>> generator.generate(size=(2, 6))
        tensor([[...]])
    """

    def __init__(
        self,
        std: BaseTensorGenerator | dict,
        max_value: BaseTensorGenerator | dict,
    ) -> None:
        super().__init__()
        self._std = setup_tensor_generator(std)
        self._max_value = setup_tensor_generator(max_value)

    def __repr__(self) -> str:
        args = str_indent(str_mapping({"std": self._std, "max_value": self._max_value}))
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def generate(self, size: tuple[int, ...], rng: Generator | None = None) -> Tensor:
        return trunc_half_normal(
            std=self._std.generate(size=size, rng=rng),
            max_value=self._max_value.generate(size=size, rng=rng),
            generator=rng,
        )



from __future__ import annotations

__all__ = ["TimeSequenceGenerator"]

from redcat import BatchedTensorSeq
from torch import Generator

from startorch.sequence.constant import ConstantSequenceGenerator
from startorch.sequence.exponential import ExponentialSequenceGenerator
from startorch.sequence.math import CumsumSequenceGenerator
from startorch.sequence.poisson import RandPoissonSequenceGenerator
from startorch.sequence.sort import SortSequenceGenerator
from startorch.sequence.uniform import RandUniformSequenceGenerator
from startorch.sequence.wrapper import BaseWrapperSequenceGenerator


class TimeSequenceGenerator(BaseWrapperSequenceGenerator):
    r"""Implements a sequence generator to generate time sequences.

    The time is represented as a float value where the unit in the
    second:

       - ``1.2`` -> ``00:00:01.200``
       - ``61.2`` -> ``00:01:01.200``
       - ``3661.2`` -> ``01:01:01.200``

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from startorch.sequence import RandUniform, Time
        >>> generator = Time(RandUniform())
        >>> generator
        TimeSequenceGenerator(
          (sequence): RandUniformSequenceGenerator(low=0.0, high=1.0, feature_size=(1,))
        )
        >>> generator.generate(seq_len=12, batch_size=4)
        tensor([[...]], batch_dim=0, seq_dim=1)
    """

    def generate(
        self, seq_len: int, batch_size: int = 1, rng: Generator | None = None
    ) -> BatchedTensorSeq:
        return self._generator.generate(seq_len=seq_len, batch_size=batch_size, rng=rng)

    @classmethod
    def create_exponential_constant_time_diff(cls, rate: float = 1.0) -> TimeSequenceGenerator:
        r"""Creates a time sequence generator where the time difference
        between two consecutive steps is constant and is sampled from an
        exponential distribution.

        Args:
        ----
            rate (float, optional): Specifies the rate of the
                exponential distribution. Default: ``1.0``

        Returns:
        -------
            ``TimeSequenceGenerator``: A time sequence generator where
                the time difference between two consecutive steps is
                constant and is sampled from an exponential
                distribution.

        Example usage:

        .. code-block:: pycon

            >>> import torch
            >>> from startorch.sequence import RandUniform, Time
            >>> generator = Time.create_exponential_constant_time_diff()
            >>> generator
            TimeSequenceGenerator(
              (sequence): CumsumSequenceGenerator(
                  (sequence): ConstantSequenceGenerator(
                      (sequence): ExponentialSequenceGenerator(
                          (rate): ConstantSequenceGenerator(
                              (sequence): RandUniformSequenceGenerator(low=1.0, high=1.0, feature_size=(1,))
                            )
                        )
                    )
                )
            )
            >>> generator.generate(seq_len=12, batch_size=4)
            tensor([[...]], batch_dim=0, seq_dim=1)
        """
        return cls(
            CumsumSequenceGenerator(
                ConstantSequenceGenerator(
                    ExponentialSequenceGenerator.create_uniform_rate(
                        min_rate=rate,
                        max_rate=rate,
                        feature_size=1,
                    )
                ),
            ),
        )

    @classmethod
    def create_exponential_time_diff(cls, rate: float = 1.0) -> TimeSequenceGenerator:
        r"""Creates a time sequence generator where the time difference
        between two consecutive steps follows an exponential
        distribution.

        Args:
        ----
            rate (float, optional): Specifies the rate of the
                exponential distribution. Default: ``1.0``

        Returns:
        -------
            ``TimeSequenceGenerator``: A time sequence generator where the
                time difference between two consecutive steps follows
                an exponential distribution.

        Example usage:

        .. code-block:: pycon

            >>> import torch
            >>> from startorch.sequence import RandUniform, Time
            >>> generator = Time.create_exponential_time_diff()
            >>> generator
            TimeSequenceGenerator(
              (sequence): CumsumSequenceGenerator(
                  (sequence): ExponentialSequenceGenerator(
                      (rate): ConstantSequenceGenerator(
                          (sequence): RandUniformSequenceGenerator(low=1.0, high=1.0, feature_size=(1,))
                        )
                    )
                )
            )
            >>> generator.generate(seq_len=12, batch_size=4)
            tensor([[...]], batch_dim=0, seq_dim=1)
        """
        return cls(
            CumsumSequenceGenerator(
                ExponentialSequenceGenerator.create_uniform_rate(
                    min_rate=rate,
                    max_rate=rate,
                    feature_size=1,
                ),
            ),
        )

    @classmethod
    def create_poisson_constant_time_diff(cls, rate: float = 1.0) -> TimeSequenceGenerator:
        r"""Creates a time sequence generator where the time difference
        between two consecutive steps is constant and is sampled from a
        Poisson distribution.

        Args:
        ----
            rate (float, optional): Specifies the rate of the Poisson
                distribution. Default: ``1.0``

        Returns:
        -------
            ``TimeSequenceGenerator``: A time sequence generator where the
                time difference between two consecutive steps is
                constant and is sampled from a Poisson distribution.

        Example usage:

        .. code-block:: pycon

            >>> import torch
            >>> from startorch.sequence import RandUniform, Time
            >>> generator = Time.create_poisson_constant_time_diff()
            >>> generator
            TimeSequenceGenerator(
              (sequence): CumsumSequenceGenerator(
                  (sequence): ConstantSequenceGenerator(
                      (sequence): RandPoissonSequenceGenerator(rate=1.0, feature_size=(1,))
                    )
                )
            )
            >>> generator.generate(seq_len=12, batch_size=4)
            tensor([[...]], batch_dim=0, seq_dim=1)
        """
        return cls(
            CumsumSequenceGenerator(
                ConstantSequenceGenerator(RandPoissonSequenceGenerator(rate, feature_size=1)),
            ),
        )

    @classmethod
    def create_poisson_time_diff(cls, rate: float = 1.0) -> TimeSequenceGenerator:
        r"""Creates a time sequence generator where the time difference
        between two consecutive steps follows a Poisson distribution.

        Args:
        ----
            rate (float, optional): Specifies the rate of the
                Poisson distribution. Default: ``1.0``

        Returns:
        -------
            ``TimeSequenceGenerator``: A time sequence generator where
                the time difference between two consecutive steps
                follows a Poisson distribution.

        Example usage:

        .. code-block:: pycon

            >>> import torch
            >>> from startorch.sequence import RandUniform, Time
            >>> generator = Time.create_poisson_time_diff()
            >>> generator
            TimeSequenceGenerator(
              (sequence): CumsumSequenceGenerator(
                  (sequence): RandPoissonSequenceGenerator(rate=1.0, feature_size=(1,))
                )
            )
            >>> generator.generate(seq_len=12, batch_size=4)
            tensor([[...]], batch_dim=0, seq_dim=1)
        """
        return cls(CumsumSequenceGenerator(RandPoissonSequenceGenerator(rate, feature_size=1)))

    @classmethod
    def create_uniform_constant_time_diff(
        cls,
        min_time_diff: float = 0.0,
        max_time_diff: float = 1.0,
    ) -> TimeSequenceGenerator:
        r"""Creates a time sequence generator where the time difference
        between two consecutive steps is constant and is sampled from a
        uniform distribution.

        Args:
        ----
            min_time_diff (float, optional): Specifies the minimum
                time difference between two consecutive steps.
                Default: ``0.0``
            max_time_diff (float, optional): Specifies the maximum
                time difference between two consecutive steps.
                Default: ``1.0``

        Returns:
        -------
            ``TimeSequenceGenerator``: A time sequence generator where
                the time difference between two consecutive steps
                is constant and is sampled from a uniform
                distribution.

        Raises:
        ------
            ValueError if ``min_time_diff`` is lower than 0.

        Example usage:

        .. code-block:: pycon

            >>> import torch
            >>> from startorch.sequence import RandUniform, Time
            >>> generator = Time.create_uniform_constant_time_diff()
            >>> generator
            TimeSequenceGenerator(
              (sequence): CumsumSequenceGenerator(
                  (sequence): ConstantSequenceGenerator(
                      (sequence): RandUniformSequenceGenerator(low=0.0, high=1.0, feature_size=(1,))
                    )
                )
            )
            >>> generator.generate(seq_len=12, batch_size=4)
            tensor([[...]], batch_dim=0, seq_dim=1)
        """
        if min_time_diff < 0:
            raise ValueError(
                f"min_time_diff has to be greater or equal to 0 (received: {min_time_diff})"
            )
        return cls(
            CumsumSequenceGenerator(
                ConstantSequenceGenerator(
                    RandUniformSequenceGenerator(
                        low=min_time_diff,
                        high=max_time_diff,
                        feature_size=1,
                    )
                ),
            ),
        )

    @classmethod
    def create_uniform_time_diff(
        cls,
        min_time_diff: float = 0.0,
        max_time_diff: float = 1.0,
    ) -> TimeSequenceGenerator:
        r"""Creates a time sequence generator where the time difference
        between two consecutive steps follows a uniform distribution.

        Args:
        ----
            min_time_diff (float, optional): Specifies the minimum
                time difference between two consecutive steps.
                Default: ``0.0``
            max_time_diff (float, optional): Specifies the maximum
                time difference between two consecutive steps.
                Default: ``1.0``

        Returns:
        -------
            ``TimeSequenceGenerator``: A time sequence generator where
                the time difference between two consecutive steps
                follows a uniform distribution.

        Raises:
        ------
            ValueError if ``min_time_diff`` is lower than 0.

        Example usage:

        .. code-block:: pycon

            >>> import torch
            >>> from startorch.sequence import RandUniform, Time
            >>> generator = Time.create_uniform_time_diff()
            >>> generator
            TimeSequenceGenerator(
              (sequence): CumsumSequenceGenerator(
                  (sequence): RandUniformSequenceGenerator(low=0.0, high=1.0, feature_size=(1,))
                )
            )
            >>> generator.generate(seq_len=12, batch_size=4)
            tensor([[...]], batch_dim=0, seq_dim=1)
        """
        if min_time_diff < 0:
            raise ValueError(
                f"min_time_diff has to be greater or equal to 0 (received: {min_time_diff})"
            )
        return cls(
            CumsumSequenceGenerator(
                RandUniformSequenceGenerator(
                    low=min_time_diff,
                    high=max_time_diff,
                    feature_size=1,
                ),
            ),
        )

    @classmethod
    def create_uniform_time(
        cls,
        min_time: float = 0.0,
        max_time: float = 1.0,
    ) -> TimeSequenceGenerator:
        r"""Creates a time sequence generator where the time is sampled
        from a uniform distribution.

        Args:
        ----
            min_time (float, optional): Specifies the minimum time.
                Default: ``0.0``
            max_time (float, optional): Specifies the maximum time.
                Default: ``1.0``

        Raises:
        ------
            ValueError if ``min_time`` is lower than 0.

        Returns:
        -------
            ``TimeSequenceGenerator``: A time sequence generator where
                the time is sampled from a uniform distribution.

        Example usage:

        .. code-block:: pycon

            >>> import torch
            >>> from startorch.sequence import RandUniform, Time
            >>> generator = Time.create_uniform_time()
            >>> generator
            TimeSequenceGenerator(
              (sequence): SortSequenceGenerator(
                  (sequence): RandUniformSequenceGenerator(low=0.0, high=1.0, feature_size=(1,))
                )
            )
            >>> generator.generate(seq_len=12, batch_size=4)
            tensor([[...]], batch_dim=0, seq_dim=1)
        """
        if min_time < 0:
            raise ValueError(f"min_time has to be greater or equal to 0 (received: {min_time})")
        return cls(
            SortSequenceGenerator(
                RandUniformSequenceGenerator(
                    low=min_time,
                    high=max_time,
                    feature_size=1,
                ),
            ),
        )



__all__ = [
    "BasePeriodicTimeSeriesGenerator",
    "Repeat",
    "is_periodic_timeseries_generator_config",
    "setup_periodic_timeseries_generator",
]

from startorch.periodic.timeseries.base import (
    BasePeriodicTimeSeriesGenerator,
    is_periodic_timeseries_generator_config,
    setup_periodic_timeseries_generator,
)
from startorch.periodic.timeseries.repeat import (
    RepeatPeriodicTimeSeriesGenerator as Repeat,
)



from __future__ import annotations

__all__ = ["ArangeSequenceGenerator"]


import torch
from redcat import BatchedTensorSeq
from torch import Generator

from startorch.sequence.base import BaseSequenceGenerator
from startorch.utils.conversion import to_tuple


class ArangeSequenceGenerator(BaseSequenceGenerator):
    r"""Implements a class to generate sequence of consecutive integer
    values between ``0`` and ``seq_len-1``.

    Args:
    ----
        feature_size (tuple or list or int, optional): Specifies the
            feature size. Default: ``1``

    Example usage:

    .. code-block:: pycon

        >>> from startorch.sequence import Arange
        >>> generator = Arange(feature_size=())
        >>> generator
        ArangeSequenceGenerator(feature_size=())
        >>> generator.generate(seq_len=6, batch_size=2)
        tensor([[0, 1, 2, 3, 4, 5],
                [0, 1, 2, 3, 4, 5]], batch_dim=0, seq_dim=1)
    """

    def __init__(self, feature_size: tuple[int, ...] | list[int] | int = 1) -> None:
        super().__init__()
        self._feature_size = to_tuple(feature_size)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(feature_size={self._feature_size})"

    def generate(
        self, seq_len: int, batch_size: int = 1, rng: Generator | None = None
    ) -> BatchedTensorSeq:
        return BatchedTensorSeq(
            torch.arange(0, seq_len)
            .view(1, seq_len, *((1,) * len(self._feature_size)))
            .repeat(batch_size, 1, *self._feature_size)
        )



from __future__ import annotations

__all__ = ["TimeSeriesExampleGenerator"]


from coola.utils import str_indent, str_mapping
from redcat import BatchedTensorSeq
from torch import Generator

from startorch.example.base import BaseExampleGenerator
from startorch.tensor.base import BaseTensorGenerator, setup_tensor_generator
from startorch.timeseries.base import (
    BaseTimeSeriesGenerator,
    setup_timeseries_generator,
)


class TimeSeriesExampleGenerator(BaseExampleGenerator):
    r"""Implements an example generator to generate time series.

    Args:
    ----
        timeseries (``BaseTimeSeriesGenerator`` or dict): Specifies a
            time series generator or its configuration.
        seq_len (``BaseTensorGenerator`` or dict): Specifies
            the sequence length sampler or its configuration. This
            sampler is used to sample the sequence length at each
            batch.

    Example usage:

    .. code-block:: pycon

        >>> from startorch import example
        >>> from startorch import timeseries
        >>> from startorch.sequence import Periodic, RandUniform
        >>> from startorch.tensor import RandInt
        >>> generator = example.TimeSeries(
        ...     timeseries=timeseries.TimeSeries({"value": RandUniform(), "time": RandUniform()}),
        ...     seq_len=RandInt(2, 5),
        ... )
        >>> generator
        TimeSeriesExampleGenerator(
          (timeseries): TimeSeriesGenerator(
              (value): RandUniformSequenceGenerator(low=0.0, high=1.0, feature_size=(1,))
              (time): RandUniformSequenceGenerator(low=0.0, high=1.0, feature_size=(1,))
            )
          (seq_len): RandIntTensorGenerator(low=2, high=5)
        )
        >>> generator.generate(batch_size=10)
        BatchDict(
          (value): tensor([...], batch_dim=0, seq_dim=1)
          (time): tensor([[...]], batch_dim=0, seq_dim=1)
        )
    """

    def __init__(
        self,
        timeseries: BaseTimeSeriesGenerator | dict,
        seq_len: BaseTensorGenerator | dict,
    ) -> None:
        super().__init__()
        self._timeseries = setup_timeseries_generator(timeseries)
        self._seq_len = setup_tensor_generator(seq_len)

    def __repr__(self) -> str:
        args = str_indent(str_mapping({"timeseries": self._timeseries, "seq_len": self._seq_len}))
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def generate(self, batch_size: int = 1, rng: Generator | None = None) -> BatchedTensorSeq:
        seq_len = int(self._seq_len.generate((1,), rng=rng).item())
        return self._timeseries.generate(seq_len=seq_len, batch_size=batch_size, rng=rng)
